import logging
import os
from typing import Optional

import orcid  # type: ignore

from mavedb.lib.logging.context import logging_context, save_to_logging_context
from mavedb.view_models.orcid import OrcidUser

logger = logging.getLogger(__name__)

ORCID_CLIENT_ID = os.getenv("ORCID_CLIENT_ID")
ORCID_CLIENT_SECRET = os.getenv("ORCID_CLIENT_SECRET")
ORCID_JWT_SIGNING_PUBLIC_KEY = os.getenv("ORCID_JWT_SIGNING_PUBLIC_KEY")


def fetch_orcid_user(orcid_id: str) -> Optional[OrcidUser]:
    """
    Given an ORCID ID, fetch the ORCID user's name.
    """
    api = orcid.PublicAPI(ORCID_CLIENT_ID, ORCID_CLIENT_SECRET)
    search_token = api.get_search_token_from_orcid()
    logger.debug(msg="Attempting to fetch user from ORCID.", extra=logging_context())
    try:
        record = api.read_record_public(orcid_id, "record", search_token)
        try:
            user = OrcidUser(
                orcid_id=record["orcid-identifier"]["path"],
                given_name=record["person"]["name"]["given-names"]["value"],
                family_name=record["person"]["name"]["family-name"]["value"],
            )
            logger.debug(msg="Successfully fetched ORCID user.", extra=logging_context())
        except Exception:
            logger.debug(
                msg="Failed to fetch ORCID user; User exists but is name metadata not visible.", extra=logging_context()
            )

    except Exception as e:
        save_to_logging_context({"orcid_exception": str(e)})
        logger.warn(msg="Encountered an error while looking up an ORCID user's email address.", extra=logging_context())
        return None

    return user


def fetch_orcid_user_email(orcid_id: str) -> Optional[str]:
    """
    Given an ORCID ID, fetch the ORCID user's name.

    The default visibility for email addresses on ORCID is `Only me`, so MaveDB will usually not be able to get the
    address.
    """

    api = orcid.PublicAPI(ORCID_CLIENT_ID, ORCID_CLIENT_SECRET)
    search_token = api.get_search_token_from_orcid()
    email: Optional[str] = None
    logger.debug(msg="Attempting to fetch user email from ORCID.", extra=logging_context())
    try:
        record = api.read_record_public(orcid_id, "record", search_token)
        try:
            email = record["person"]["emails"]["email"][0]["email"]
            logger.debug(msg="Successfully fetched ORCID email.", extra=logging_context())
        except Exception:
            logger.debug(msg="Failed to fetch ORCID email; User exists but email not visible.", extra=logging_context())

    except Exception as e:
        save_to_logging_context({"orcid_exception": str(e)})
        logger.warn(msg="Encountered an error while looking up an ORCID user's email address.", extra=logging_context())

    return email



# import os

from typing import Any, AsyncGenerator, Generator

from arq import ArqRedis, create_pool
from cdot.hgvs.dataproviders import RESTDataProvider
from sqlalchemy.orm import Session

from mavedb.data_providers.services import cdot_rest
from mavedb.db.session import SessionLocal
from mavedb.worker.settings import RedisWorkerSettings


def get_db() -> Generator[Session, Any, None]:
    db = SessionLocal()
    db.current_user_id = None  # type: ignore
    try:
        yield db
    finally:
        db.close()


async def get_worker() -> AsyncGenerator[ArqRedis, Any]:
    redis = await create_pool(RedisWorkerSettings)
    try:
        yield redis
    finally:
        await redis.close()


def hgvs_data_provider() -> RESTDataProvider:
    return cdot_rest()



from datetime import date
from typing import Optional

from mavedb.view_models import record_type_validator, set_record_type
from mavedb.view_models.base.base import BaseModel, validator


class TargetAccessionBase(BaseModel):
    accession: str
    assembly: Optional[str]
    gene: Optional[str]

    @validator("gene", always=True)
    def check_gene_or_assembly(cls, gene, values):
        if "assembly" not in values and not gene:
            raise ValueError("either a `gene` or `assembly` is required")
        return gene


class TargetAccessionModify(TargetAccessionBase):
    # Consider some validation, ie: Accession is in our CDOT data files
    pass


class TargetAccessionCreate(TargetAccessionModify):
    pass


class TargetAccessionUpdate(TargetAccessionModify):
    pass


# Properties shared by models stored in DB
class SavedTargetAccession(TargetAccessionBase):
    record_type: str = None  # type: ignore

    _record_type_factory = record_type_validator()(set_record_type)

    class Config:
        orm_mode = True
        arbitrary_types_allowed = True


# Properties to return to non-admin clients
class TargetAccession(SavedTargetAccession):
    pass


# Properties to return to admin clients
class AdminTargetAccession(SavedTargetAccession):
    creation_date: date
    modification_date: date



from datetime import date
from typing import TYPE_CHECKING

from sqlalchemy import Column, Date, Enum, ForeignKey, Integer, String
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import Mapped, backref, relationship

from mavedb.db.base import Base
from mavedb.models.enums.target_category import TargetCategory
from mavedb.models.score_set import ScoreSet
from mavedb.models.target_accession import TargetAccession
from mavedb.models.target_sequence import TargetSequence

if TYPE_CHECKING:
    from mavedb.models.ensembl_offset import EnsemblOffset
    from mavedb.models.refseq_offset import RefseqOffset
    from mavedb.models.uniprot_offset import UniprotOffset

# TODO Reformat code without removing dependencies whose use is not detected.


class TargetGene(Base):
    __tablename__ = "target_genes"

    id = Column(Integer, primary_key=True)

    name = Column(String, nullable=False)
    category = Column(
        Enum(TargetCategory, create_constraint=True, length=32, native_enum=False, validate_strings=True),
        nullable=False,
    )

    score_set_id = Column("scoreset_id", Integer, ForeignKey("scoresets.id"), index=True, nullable=False)
    score_set: Mapped[ScoreSet] = relationship(back_populates="target_genes", single_parent=True, uselist=True)

    target_sequence_id = Column(Integer, ForeignKey("target_sequences.id"), index=True, nullable=True)
    accession_id = Column(Integer, ForeignKey("target_accessions.id"), index=True, nullable=True)
    target_sequence: Mapped[TargetSequence] = relationship(
        back_populates="target_genes", cascade="all,delete-orphan", single_parent=True
    )
    target_accession: Mapped[TargetAccession] = relationship(
        "TargetAccession",
        backref=backref("target_genes", single_parent=True, uselist=True),
        cascade="all,delete-orphan",
        single_parent=True,
    )

    pre_mapped_metadata: Mapped[JSONB] = Column("pre_mapped_metadata", JSONB, nullable=True)
    post_mapped_metadata: Mapped[JSONB] = Column("post_mapped_metadata", JSONB, nullable=True)

    creation_date = Column(Date, nullable=False, default=date.today)
    modification_date = Column(Date, nullable=False, default=date.today, onupdate=date.today)

    ensembl_offset: Mapped["EnsemblOffset"] = relationship(back_populates="target_gene", cascade="all, delete-orphan")
    refseq_offset: Mapped["RefseqOffset"] = relationship(back_populates="target_gene", cascade="all, delete-orphan")
    uniprot_offset: Mapped["UniprotOffset"] = relationship(back_populates="target_gene", cascade="all, delete-orphan")



from fastapi import APIRouter, Depends
from fastapi.exceptions import HTTPException
from sqlalchemy import func
from sqlalchemy.orm import Session

from mavedb import deps
from mavedb.lib.keywords import search_keyword as _search_keyword
from mavedb.models.controlled_keyword import ControlledKeyword
from mavedb.view_models import keyword

router = APIRouter(
    prefix="/api/v1/controlled-keywords", tags=["controlled-keywords"], responses={404: {"description": "Not found"}}
)


@router.get(
    "/{key}",
    status_code=200,
    response_model=list[keyword.Keyword],
    responses={404: {}},
    response_model_exclude_none=True,
)
def fetch_keywords_by_key(
    *,
    key: str,
    db: Session = Depends(deps.get_db),
) -> list[ControlledKeyword]:
    """
    Fetch keywords by category.
    """
    lower_key = key.lower()
    items = (
        db.query(ControlledKeyword)
        .filter(func.lower(ControlledKeyword.key) == lower_key)
        .order_by(ControlledKeyword.value)
        .all()
    )
    if not items:
        raise HTTPException(status_code=404, detail=f"Controlled keywords with key {key} not found")
    return items


@router.post("/search/{key}/{value}", status_code=200, response_model=keyword.Keyword)
def search_keyword_by_key_and_value(key: str, value: str, db: Session = Depends(deps.get_db)) -> ControlledKeyword:
    """
    Search keywords.
    """
    return _search_keyword(db, key, value)



from datetime import date

from sqlalchemy import Boolean, Column, Date, ForeignKey, Integer, String
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import Mapped, backref, relationship

from mavedb.db.base import Base

from .variant import Variant


class MappedVariant(Base):
    __tablename__ = "mapped_variants"

    id = Column(Integer, primary_key=True)

    pre_mapped = Column(JSONB(none_as_null=True), nullable=True)
    post_mapped = Column(JSONB(none_as_null=True), nullable=True)
    vrs_version = Column(String, nullable=True)
    error_message = Column(String, nullable=True)
    modification_date = Column(Date, nullable=False, default=date.today, onupdate=date.today)
    mapped_date = Column(Date, nullable=False)
    mapping_api_version = Column(String, nullable=False)
    current = Column(Boolean, nullable=False)

    variant_id = Column(Integer, ForeignKey("variants.id"), index=True, nullable=False)
    variant: Mapped[Variant] = relationship("Variant", backref=backref("mapped_variants", cascade="all,delete-orphan"))



from datetime import date
from typing import TYPE_CHECKING, List, Optional

from sqlalchemy import Boolean, Column, Date, Enum, ForeignKey, Integer, String
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.associationproxy import AssociationProxy, association_proxy
from sqlalchemy.orm import Mapped, relationship
from sqlalchemy.schema import Table

import mavedb.models.score_set_publication_identifier
from mavedb.db.base import Base
from mavedb.models.collection_association import collection_score_sets_association_table
from mavedb.models.contributor import Contributor
from mavedb.models.doi_identifier import DoiIdentifier
from mavedb.models.enums.mapping_state import MappingState
from mavedb.models.enums.processing_state import ProcessingState
from mavedb.models.experiment import Experiment
from mavedb.models.legacy_keyword import LegacyKeyword
from mavedb.models.license import License
from mavedb.models.publication_identifier import PublicationIdentifier
from mavedb.models.user import User

if TYPE_CHECKING:
    from mavedb.models.collection import Collection
    from mavedb.models.target_gene import TargetGene
    from mavedb.models.variant import Variant

# from .raw_read_identifier import SraIdentifier
from mavedb.lib.temp_urns import generate_temp_urn

# TODO Reformat code without removing dependencies whose use is not detected.

score_sets_contributors_association_table = Table(
    "scoreset_contributors",
    Base.metadata,
    Column("scoreset_id", ForeignKey("scoresets.id"), primary_key=True),
    Column("contributor_id", ForeignKey("contributors.id"), primary_key=True),
)


score_sets_doi_identifiers_association_table = Table(
    "scoreset_doi_identifiers",
    Base.metadata,
    Column("scoreset_id", ForeignKey("scoresets.id"), primary_key=True),
    Column("doi_identifier_id", ForeignKey("doi_identifiers.id"), primary_key=True),
)


score_sets_legacy_keywords_association_table = Table(
    "scoreset_keywords",
    Base.metadata,
    Column("scoreset_id", ForeignKey("scoresets.id"), primary_key=True),
    Column("keyword_id", ForeignKey("keywords.id"), primary_key=True),
)


score_sets_meta_analysis_score_sets_association_table = Table(
    "scoreset_meta_analysis_sources",
    Base.metadata,
    Column("source_scoreset_id", ForeignKey("scoresets.id"), primary_key=True),
    Column("meta_analysis_scoreset_id", ForeignKey("scoresets.id"), primary_key=True),
)


# score_sets_sra_identifiers_association_table = Table(
score_sets_raw_read_identifiers_association_table = Table(
    "scoreset_sra_identifiers",
    Base.metadata,
    Column("scoreset_id", ForeignKey("scoresets.id"), primary_key=True),
    Column("sra_identifier_id", ForeignKey("sra_identifiers.id"), primary_key=True),
)


class ScoreSet(Base):
    __tablename__ = "scoresets"

    id = Column(Integer, primary_key=True)

    urn = Column(String(64), default=generate_temp_urn, index=True, nullable=True, unique=True)
    title = Column(String, nullable=False)
    method_text = Column(String, nullable=False)
    abstract_text = Column(String, nullable=False)
    short_description = Column(String, nullable=False)
    extra_metadata = Column(JSONB, nullable=False)
    dataset_columns = Column(JSONB, nullable=False, default={})
    external_links = Column(JSONB, nullable=False, default={})

    normalised = Column(Boolean, nullable=False, default=False)
    private = Column(Boolean, nullable=False, default=True)
    approved = Column(Boolean, nullable=False, default=False)
    published_date = Column(Date, nullable=True)
    processing_state = Column(
        Enum(
            ProcessingState,
            create_constraint=True,
            length=32,
            native_enum=False,
            validate_strings=True,
        ),
        nullable=True,
    )
    processing_errors = Column(JSONB, nullable=True)
    data_usage_policy = Column(String, nullable=True)

    num_variants = Column(Integer, nullable=False, default=0)
    variants: Mapped[list["Variant"]] = relationship(back_populates="score_set", cascade="all, delete-orphan")

    mapping_state = Column(
        Enum(
            MappingState,
            create_constraint=True,
            length=32,
            native_enum=False,
            validate_strings=True,
        ),
        nullable=True,
    )
    mapping_errors = Column(JSONB, nullable=True)

    experiment_id = Column(Integer, ForeignKey("experiments.id"), index=True, nullable=False)
    experiment: Mapped["Experiment"] = relationship(back_populates="score_sets")

    # TODO Standardize on US or GB spelling for licenc/se.
    licence_id = Column(Integer, ForeignKey("licenses.id"), index=True, nullable=False)
    license: Mapped["License"] = relationship("License")
    superseded_score_set_id = Column("replaces_id", Integer, ForeignKey("scoresets.id"), index=True, nullable=True)
    superseded_score_set: Mapped[Optional["ScoreSet"]] = relationship(
        "ScoreSet",
        uselist=False,
        foreign_keys="ScoreSet.superseded_score_set_id",
        remote_side=[id],
    )
    superseding_score_set: Mapped[Optional["ScoreSet"]] = relationship(
        "ScoreSet", uselist=False, back_populates="superseded_score_set"
    )

    created_by_id = Column(Integer, ForeignKey("users.id"), index=True, nullable=True)
    created_by: Mapped["User"] = relationship("User", foreign_keys="ScoreSet.created_by_id")
    modified_by_id = Column(Integer, ForeignKey("users.id"), index=True, nullable=True)
    modified_by: Mapped["User"] = relationship("User", foreign_keys="ScoreSet.modified_by_id")
    creation_date = Column(Date, nullable=False, default=date.today)
    modification_date = Column(Date, nullable=False, default=date.today, onupdate=date.today)

    legacy_keyword_objs: Mapped[list["LegacyKeyword"]] = relationship(
        "LegacyKeyword",
        secondary=score_sets_legacy_keywords_association_table,
        backref="score_sets",
    )
    contributors: Mapped[list["Contributor"]] = relationship(
        "Contributor",
        secondary=score_sets_contributors_association_table,
        backref="score_sets",
    )
    doi_identifiers: Mapped[list["DoiIdentifier"]] = relationship(
        "DoiIdentifier",
        secondary=score_sets_doi_identifiers_association_table,
        backref="score_sets",
    )
    publication_identifier_associations: Mapped[
        list[mavedb.models.score_set_publication_identifier.ScoreSetPublicationIdentifierAssociation]
    ] = relationship(
        "ScoreSetPublicationIdentifierAssociation",
        back_populates="score_set",
        cascade="all, delete-orphan",
    )
    publication_identifiers: AssociationProxy[List[PublicationIdentifier]] = association_proxy(
        "publication_identifier_associations",
        "publication",
        creator=lambda p: mavedb.models.score_set_publication_identifier.ScoreSetPublicationIdentifierAssociation(
            publication=p, primary=p.primary
        ),
    )
    # sra_identifiers = relationship('SraIdentifier', secondary=score_sets_sra_identifiers_association_table, backref='score_sets')
    # raw_read_identifiers = relationship('RawReadIdentifier', secondary=score_sets_raw_read_identifiers_association_table,backref='score_sets')
    meta_analyzes_score_sets: Mapped[list["ScoreSet"]] = relationship(
        "ScoreSet",
        secondary=score_sets_meta_analysis_score_sets_association_table,
        primaryjoin=(score_sets_meta_analysis_score_sets_association_table.c.meta_analysis_scoreset_id == id),
        secondaryjoin=(score_sets_meta_analysis_score_sets_association_table.c.source_scoreset_id == id),
        backref="meta_analyzed_by_score_sets",
    )

    target_genes: Mapped[List["TargetGene"]] = relationship(back_populates="score_set", cascade="all, delete-orphan")
    score_ranges = Column(JSONB, nullable=True)
    score_calibrations = Column(JSONB, nullable=True)

    collections: Mapped[list["Collection"]] = relationship(
        "Collection",
        secondary=collection_score_sets_association_table,
        back_populates="score_sets",
    )
    official_collections: Mapped[list["Collection"]] = relationship(
        "Collection",
        secondary=collection_score_sets_association_table,
        secondaryjoin="and_(collection_score_sets.c.collection_id == Collection.id, Collection.badge_name != None)",
        back_populates="score_sets",
        viewonly=True,
    )

    # Unfortunately, we can't use association_proxy here, because in spite of what the documentation seems to imply, it
    # doesn't check for a pre-existing keyword with the same text.
    # keywords = association_proxy('keyword_objs', 'text', creator=lambda text: Keyword(text=text))

    # _updated_keywords: list[str] = None
    # _updated_doi_identifiers: list[str] = None

    @property
    def legacy_keywords(self) -> list[str]:
        # if self._updated_keywords:
        #     return self._updated_keywords
        # else:
        legacy_keyword_objs = self.legacy_keyword_objs or []  # getattr(self, 'keyword_objs', [])
        return [
            legacy_keyword_obj.text for legacy_keyword_obj in legacy_keyword_objs if legacy_keyword_obj.text is not None
        ]

    async def set_legacy_keywords(self, db, keywords: list[str]):
        self.keyword_objs = [await self._find_or_create_legacy_keyword(db, text) for text in keywords]

    # See https://gist.github.com/tachyondecay/e0fe90c074d6b6707d8f1b0b1dcc8e3a
    # @keywords.setter
    # async def set_keywords(self, db, keywords: list[str]):
    #     self._keyword_objs = [await self._find_or_create_keyword(text) for text in keywords]

    async def _find_or_create_legacy_keyword(self, db, keyword_text):
        keyword_obj = db.query(LegacyKeyword).filter(LegacyKeyword.text == keyword_text).one_or_none()
        if not keyword_obj:
            keyword_obj = LegacyKeyword(text=keyword_text)
        return keyword_obj



from typing import Any

from fastapi import APIRouter

from mavedb import __project__, __version__
from mavedb.view_models import api_version

router = APIRouter(prefix="/api/v1/api", tags=["api information"], responses={404: {"description": "Not found"}})


@router.get("/version", status_code=200, response_model=api_version.ApiVersion, responses={404: {}})
def show_version() -> Any:
    """
    Describe the API version.
    """

    return api_version.ApiVersion(name=__project__, version=__version__)



import logging
from enum import Enum
from typing import Optional, Union

from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy.orm import Session

from mavedb import deps
from mavedb.lib.authentication import UserData, get_current_user
from mavedb.lib.logging import LoggedRoute
from mavedb.lib.logging.context import logging_context, save_to_logging_context
from mavedb.lib.permissions import Action, has_permission
from mavedb.models.collection import Collection
from mavedb.models.experiment import Experiment
from mavedb.models.experiment_set import ExperimentSet
from mavedb.models.score_set import ScoreSet

router = APIRouter(
    prefix="/api/v1/permissions",
    tags=["permissions"],
    responses={404: {"description": "Not found"}},
    route_class=LoggedRoute,
)

logger = logging.getLogger(__name__)


class ModelName(str, Enum):
    collection = "collection"
    experiment = "experiment"
    experiment_set = "experiment-set"
    score_set = "score-set"


@router.get(
    "/user-is-permitted/{model_name}/{urn}/{action}",
    status_code=200,
    response_model=bool,
)
async def check_permission(
    *,
    model_name: ModelName,
    urn: str,
    action: Action,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(get_current_user),
) -> bool:
    """
    Check whether users have authorizations in adding/editing/deleting/publishing experiment or score set.
    """
    save_to_logging_context({"requested_resource": urn})

    item: Optional[Union[Collection, ExperimentSet, Experiment, ScoreSet]] = None

    if model_name == ModelName.experiment_set:
        item = db.query(ExperimentSet).filter(ExperimentSet.urn == urn).one_or_none()
    elif model_name == ModelName.experiment:
        item = db.query(Experiment).filter(Experiment.urn == urn).one_or_none()
    elif model_name == ModelName.score_set:
        item = db.query(ScoreSet).filter(ScoreSet.urn == urn).one_or_none()
    elif model_name == ModelName.collection:
        item = db.query(Collection).filter(Collection.urn == urn).one_or_none()

    if item:
        permission = has_permission(user_data, item, action).permitted
        return permission
    else:
        logger.debug(msg="The requested resources does not exist.", extra=logging_context())
        raise HTTPException(status_code=404, detail=f"{model_name.value} with URN '{urn}' not found")



import logging
from typing import Any

from pythonjsonlogger.jsonlogger import JsonFormatter


class MavedbJsonFormatter(JsonFormatter):
    def add_fields(self, log_record: dict[str, Any], record: logging.LogRecord, message_dict: dict[str, Any]) -> None:
        """
        Override JsonFormatter to add level and filepath to emitted messages.
        """
        message_dict["level"] = record.levelname
        message_dict["filepath"] = record.pathname

        super(MavedbJsonFormatter, self).add_fields(log_record, record, message_dict)



from sqlalchemy import Column, Integer, String

from mavedb.db.base import Base


class Contributor(Base):
    __tablename__ = "contributors"

    id = Column(Integer, primary_key=True)
    orcid_id = Column(String, index=True, nullable=False, unique=True)
    given_name = Column(String, nullable=True)
    family_name = Column(String, nullable=True)



import logging
from typing import Optional

from fastapi import Depends, HTTPException
from starlette import status

from mavedb.lib.authentication import UserData, get_current_user
from mavedb.lib.logging.context import logging_context, save_to_logging_context
from mavedb.models.enums.user_role import UserRole

logger = logging.getLogger(__name__)


####################################################################################################
# Main authorization methods
####################################################################################################


async def require_current_user(
    user_data: Optional[UserData] = Depends(get_current_user),
) -> UserData:
    if user_data is None:
        logger.info(msg="Non-authenticated user attempted to access protected route.", extra=logging_context())
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Could not validate credentials",
        )

    return user_data


async def require_current_user_with_email(
    user_data: UserData = Depends(require_current_user),
) -> UserData:
    # Both empty strings and NoneType values should raise an exception.
    if not user_data.user.email:
        logger.info(
            msg="User attempted to access email protected route without a valid email.", extra=logging_context()
        )
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="There must be an email address associated with your account to use this feature.",
        )
    return user_data


class RoleRequirer:
    def __init__(self, roles: list[UserRole]):
        self.roles = roles

    async def __call__(self, user_data: UserData = Depends(require_current_user)) -> UserData:
        save_to_logging_context({"required_roles": [role.name for role in self.roles]})
        if not any(role in self.roles for role in user_data.active_roles):
            logger.info(
                msg="User attempted to access role protected route without a required role.", extra=logging_context()
            )
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="You are not authorized to use this feature",
            )

        return user_data


async def require_role(roles: list[UserRole], user_data: UserData = Depends(require_current_user)) -> UserData:
    save_to_logging_context({"required_roles": [role.name for role in roles]})
    if not any(role.name in roles for role in user_data.active_roles):
        logger.info(
            msg="User attempted to access role protected route without a required role.", extra=logging_context()
        )
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="You are not authorized to use this feature",
        )

    return user_data



from enum import Enum
from typing import Any, List

from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy import func, or_
from sqlalchemy.exc import MultipleResultsFound
from sqlalchemy.orm import Session
from starlette.convertors import Convertor, register_url_convertor

from mavedb import deps
from mavedb.lib.identifiers import find_generic_article
from mavedb.lib.validation.constants.publication import valid_dbnames
from mavedb.models.publication_identifier import PublicationIdentifier
from mavedb.view_models import publication_identifier
from mavedb.view_models.search import TextSearch

# I don't think we can escape the type: ignore hint here on a dynamically created enumerated type.
PublicationDatabases = Enum("PublicationDataBases", ((x, x) for x in valid_dbnames))  # type: ignore


router = APIRouter(
    prefix="/api/v1/publication-identifiers",
    tags=["publication identifiers"],
    responses={404: {"description": "Not found"}},
)


@router.get("/", status_code=200, response_model=list[publication_identifier.PublicationIdentifier])
def list_publications(*, db: Session = Depends(deps.get_db)) -> Any:
    """
    List stored all stored publications.
    """
    items = db.query(PublicationIdentifier).all()
    return items


# See https://github.com/tiangolo/fastapi/discussions/7328, which describes that slashes are currently un-escapable in FastAPI/Starlette
# implementations.
#
# Workaround here by defining a custom type convertor (see https://www.starlette.io/routing/#path-parameters) whose
# RegEx match matches our accepted publication identifiers. When the API is queried and the string matches the convertor, we enter the
# route on which the convertor captured a match, use the match as our variable of interest, and resolve the route. By capturing the match,
# we retain the ability to add routes such as /db_name/identifier or /identifier/title since we can now match to these unescapable slashes
# without matching the rest of the path as a path type, as the `:path` convertor would have. In the OpenAPI spec, this type will still be
# represented as a 'path'. -capodb 2024.05.30
class PublicationIdentifierConverter(Convertor):
    # RegEx structure: "DOI | PubMed | bioRxiv | medRxiv"
    regex = "10[.][0-9]{4,9}\/[-._;()\/:A-z0-9]+|[0-9]{0,8}|[0-9]{4}[.][0-9]{2}[.][0-9]{2}[.][0-9]{6}|[0-9]{4}[.][0-9]{2}[.][0-9]{2}[.][0-9]{8}"

    def convert(self, value: str) -> str:
        return value

    def to_string(self, value: str) -> str:
        return str(value)


register_url_convertor("publication", PublicationIdentifierConverter())


@router.get(
    "/{identifier:publication}",
    status_code=200,
    response_model=publication_identifier.PublicationIdentifier,
    responses={404: {}},
)
def fetch_publication_by_identifier(*, identifier: str, db: Session = Depends(deps.get_db)) -> PublicationIdentifier:
    """
    Fetch a single publication by identifier.
    """
    try:
        item = db.query(PublicationIdentifier).filter(PublicationIdentifier.identifier == identifier).one_or_none()
    except MultipleResultsFound:
        raise HTTPException(
            status_code=500,
            detail=f"Multiple publications with identifier {identifier} were found.",
        )

    if not item:
        raise HTTPException(status_code=404, detail=f"No publication with identifier {identifier} were found.")
    return item


@router.get(
    "/{db_name:str}/{identifier:publication}",
    status_code=200,
    response_model=publication_identifier.PublicationIdentifier,
    responses={404: {}},
)
def fetch_publication_by_dbname_and_identifier(
    *,
    db_name: PublicationDatabases,
    identifier: str,
    db: Session = Depends(deps.get_db),
) -> PublicationIdentifier:
    """
    Fetch a single publication by db name and identifier.
    """
    try:
        item = (
            db.query(PublicationIdentifier)
            .filter(PublicationIdentifier.identifier == identifier)
            .filter(PublicationIdentifier.db_name == db_name.name)
            .one_or_none()
        )
    except MultipleResultsFound:
        raise HTTPException(
            status_code=500,
            detail=f"Multiple publications with identifier {identifier} and database name {db_name.name} were found.",
        )
    if not item:
        raise HTTPException(
            status_code=404,
            detail=f"No publication with identifier {identifier} and database name {db_name.name} were found.",
        )
    return item


@router.get("/journals", status_code=200, response_model=list[str], responses={404: {}})
def list_publication_journal_names(*, db: Session = Depends(deps.get_db)) -> Any:
    """
    List distinct journal names, in alphabetical order.
    """

    items = db.query(PublicationIdentifier).all()
    journals = map(lambda item: item.publication_journal, items)
    return sorted(list(set(journals)))


@router.get("/databases", status_code=200, response_model=list[str], responses={404: {}})
def list_publication_database_names(*, db: Session = Depends(deps.get_db)) -> Any:
    """
    List distinct database names, in alphabetical order.
    """

    items = db.query(PublicationIdentifier).all()
    databases = map(lambda item: item.db_name, items)
    return sorted(list(set(databases)))


@router.post("/search/identifier", status_code=200, response_model=list[publication_identifier.PublicationIdentifier])
def search_publication_identifier_identifiers(search: TextSearch, db: Session = Depends(deps.get_db)) -> Any:
    """
    Search publication identifiers via a TextSearch query.
    """

    query = db.query(PublicationIdentifier)

    if search.text and len(search.text.strip()) > 0:
        lower_search_text = search.text.strip().lower()
        query = query.filter(func.lower(PublicationIdentifier.identifier).contains(lower_search_text))
    else:
        raise HTTPException(status_code=500, detail="Search text is required")

    items = query.order_by(PublicationIdentifier.identifier).limit(50).all()
    if not items:
        items = []
    return items


@router.post("/search/doi", status_code=200, response_model=list[publication_identifier.PublicationIdentifier])
def search_publication_identifier_dois(search: TextSearch, db: Session = Depends(deps.get_db)) -> Any:
    """
    Search publication DOIs via a TextSearch query.
    """

    query = db.query(PublicationIdentifier)

    if search.text and len(search.text.strip()) > 0:
        lower_search_text = search.text.strip().lower()
        query = query.filter(func.lower(PublicationIdentifier.doi).contains(lower_search_text))
    else:
        raise HTTPException(status_code=500, detail="Search text is required")

    items = query.order_by(PublicationIdentifier.doi).limit(50).all()
    if not items:
        items = []
    return items


@router.post("/search", status_code=200, response_model=list[publication_identifier.PublicationIdentifier])
def search_publication_identifiers(search: TextSearch, db: Session = Depends(deps.get_db)) -> Any:
    """
    Search publication identifiers via a TextSearch query, returning substring matches on DOI and Identifier.
    """

    query = db.query(PublicationIdentifier)

    if search.text and len(search.text.strip()) > 0:
        lower_search_text = search.text.strip().lower()
        query = query.filter(
            or_(
                func.lower(PublicationIdentifier.identifier).contains(lower_search_text),
                func.lower(PublicationIdentifier.doi).contains(lower_search_text),
            )
        )
    else:
        raise HTTPException(status_code=500, detail="Search text is required")

    items = query.order_by(PublicationIdentifier.identifier).limit(50).all()
    if not items:
        items = []
    return items


@router.get(
    "/search/{identifier}",
    status_code=200,
    response_model=publication_identifier.PublicationIdentifier,
    responses={404: {}, 500: {}},
)
async def search_publications_by_identifier(*, identifier: str, db: Session = Depends(deps.get_db)) -> Any:
    """
    Search publication identifiers via their identifier.
    """
    query = db.query(PublicationIdentifier).filter(PublicationIdentifier.identifier == identifier).all()

    if not query:
        raise HTTPException(status_code=404, detail=f"No publications with identifier {identifier} were found.")
    return query


@router.get(
    "/search/{db_name}/{identifier}",
    status_code=200,
    response_model=list[publication_identifier.PublicationIdentifier],
    responses={404: {}, 500: {}},
)
async def search_publications_by_identifier_and_db(
    *,
    identifier: str,
    db_name: str,
    db: Session = Depends(deps.get_db),
) -> Any:
    """
    Search all of the publication identifiers via their identifier and database.
    """
    query = (
        db.query(PublicationIdentifier)
        .filter(PublicationIdentifier.identifier == identifier and PublicationIdentifier.db_name == db_name)
        .all()
    )

    if not query:
        raise HTTPException(
            status_code=404,
            detail=f"No publications with identifier {identifier} and database name {db_name} were found.",
        )
    return query


@router.post(
    "/search-external", status_code=200, response_model=List[publication_identifier.ExternalPublicationIdentifier]
)
async def search_external_publication_identifiers(search: TextSearch, db: Session = Depends(deps.get_db)) -> Any:
    """
    Search external publication identifiers via a TextSearch query.
    Technically, this should be some sort of accepted publication identifier.
    """

    if search.text and len(search.text.strip()) > 0:
        lower_search_text = search.text.strip().lower()
        items = await find_generic_article(db, lower_search_text)
    else:
        raise HTTPException(status_code=500, detail="Search text is required")

    if not any(items.values()):
        raise HTTPException(status_code=404, detail="No publications matched the provided search text")

    return [item for item in items.values() if item is not None]



import logging
from typing import Any

from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy.orm import Session
from starlette.convertors import Convertor, register_url_convertor

from mavedb import deps
from mavedb.lib.authentication import UserData
from mavedb.lib.authorization import RoleRequirer, require_current_user
from mavedb.lib.logging import LoggedRoute
from mavedb.lib.logging.context import logging_context, save_to_logging_context
from mavedb.lib.permissions import Action, assert_permission
from mavedb.models.enums.user_role import UserRole
from mavedb.models.user import User
from mavedb.view_models import user

router = APIRouter(
    prefix="/api/v1",
    tags=["access keys"],
    responses={404: {"description": "Not found"}},
    route_class=LoggedRoute,
)

logger = logging.getLogger(__name__)


# Define custom type convertor (see https://www.starlette.io/routing/#path-parameters)
# in order to recognize user lookup id as an int or orcid id, and call the appropriate function
class OrcidIdConverter(Convertor):
    regex = "\d{4}-\d{4}-\d{4}-(\d{4}|\d{3}X)"

    def convert(self, value: str) -> str:
        return value

    def to_string(self, value: str) -> str:
        return str(value)


register_url_convertor("orcid_id", OrcidIdConverter())


# Trailing slash is deliberate
@router.get("/users/", status_code=200, response_model=list[user.AdminUser], responses={404: {}})
async def list_users(
    *,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(RoleRequirer([UserRole.admin])),
) -> Any:
    """
    List users.
    """
    items = db.query(User).order_by(User.username).all()
    return items


@router.get(
    "/users/me",
    status_code=200,
    response_model=user.CurrentUser,
    responses={404: {}, 500: {}},
)
async def show_me(*, user_data: UserData = Depends(require_current_user)) -> Any:
    """
    Return the current user.
    """
    return user_data.user


@router.get(
    "/users/{id:int}",
    status_code=200,
    response_model=user.AdminUser,
    responses={404: {}, 500: {}},
)
async def show_user_admin(
    *,
    id: int,
    user_data: UserData = Depends(RoleRequirer([UserRole.admin])),
    db: Session = Depends(deps.get_db),
) -> Any:
    """
    Fetch a single user by ID. Returns admin view of requested user.
    """
    save_to_logging_context({"requested_user": id})
    item = db.query(User).filter(User.id == id).one_or_none()
    if not item:
        logger.warning(
            msg="Could not show user; Requested user does not exist.",
            extra=logging_context(),
        )
        raise HTTPException(status_code=404, detail=f"User with ID {id} not found")

    # moving toward always accessing permissions module, even though this function does already require admin role to access
    assert_permission(user_data, item, Action.READ)
    return item


@router.get(
    "/users/{orcid_id:orcid_id}",
    status_code=200,
    response_model=user.User,
    responses={404: {}, 500: {}},
)
async def show_user(
    *,
    orcid_id: str,
    user_data: UserData = Depends(require_current_user),
    db: Session = Depends(deps.get_db),
) -> Any:
    """
    Fetch a single user by Orcid ID. Returns limited view of user.
    """
    save_to_logging_context({"requested_user": orcid_id})

    item = db.query(User).filter(User.username == orcid_id).one_or_none()
    if not item:
        logger.warning(
            msg="Could not show user; Requested user does not exist.",
            extra=logging_context(),
        )
        raise HTTPException(status_code=404, detail=f"User with ID {orcid_id} not found")

    # moving toward always accessing permissions module, even though this function does already require existing user in order to access
    assert_permission(user_data, item, Action.LOOKUP)
    return item


@router.put(
    "/users/me",
    status_code=200,
    response_model=user.CurrentUser,
    responses={404: {}, 500: {}},
)
async def update_me(
    *,
    user_update: user.CurrentUserUpdate,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user),
) -> Any:
    """
    Update the current user.
    """
    current_user = user_data.user
    assert_permission(user_data, current_user, Action.UPDATE)
    current_user.email = user_update.email
    current_user.is_first_login = False
    db.add(current_user)
    db.commit()
    db.refresh(current_user)
    return current_user


@router.put(
    "/users/me/has-logged-in",
    status_code=200,
    response_model=user.CurrentUser,
    responses={404: {}, 500: {}},
)
async def user_has_logged_in(
    *,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user),
) -> Any:
    """
    Update the current users log in state.
    """
    current_user = user_data.user
    assert_permission(user_data, current_user, Action.UPDATE)
    current_user.is_first_login = False
    db.add(current_user)
    db.commit()
    db.refresh(current_user)
    return current_user


# Double slash is deliberate.
@router.put(
    "/users//{id}",
    status_code=200,
    response_model=user.AdminUser,
    responses={404: {}, 500: {}},
)
async def update_user(
    *,
    id: int,
    item_update: user.AdminUserUpdate,
    db: Session = Depends(deps.get_db),
    user_data: UserData = Depends(require_current_user),
) -> Any:
    """
    Update a user.
    """
    save_to_logging_context({"requested_user": id})
    item = db.query(User).filter(User.id == id).one_or_none()
    if not item:
        logger.warning(
            msg="Could not update user; Requested user does not exist.",
            extra=logging_context(),
        )
        raise HTTPException(status_code=404, detail=f"User with id {id} not found.")

    assert_permission(user_data, item, Action.UPDATE)
    assert_permission(user_data, item, Action.ADD_ROLE)

    if item_update.first_name:
        item.first_name = item_update.first_name
    if item_update.last_name:
        item.last_name = item_update.last_name
    if item_update.email:
        item.email = item_update.email
    if item_update.roles:
        await item.set_roles(db, item_update.roles)

    db.add(item)
    db.commit()
    db.refresh(item)
    return item



from datetime import date
from typing import Optional

from fqfa import infer_sequence_type

from mavedb.lib.validation import target
from mavedb.lib.validation.exceptions import ValidationError
from mavedb.view_models import record_type_validator, set_record_type
from mavedb.view_models.base.base import BaseModel, validator
from mavedb.view_models.taxonomy import AdminTaxonomy, SavedTaxonomy, Taxonomy, TaxonomyCreate


def sanitize_target_sequence_label(label: str):
    return label.strip().replace(" ", "_")


class TargetSequenceBase(BaseModel):
    sequence_type: str
    sequence: str
    label: Optional[str]


class TargetSequenceModify(TargetSequenceBase):
    @validator("sequence_type")
    def validate_category(cls, field_value, values, field, config):
        field_value = field_value.lower()
        target.validate_sequence_category(field_value)
        if field_value == "infer":
            if "sequence" in values.keys():
                field_value = infer_sequence_type(values["sequence"])
            else:
                raise ValueError("sequence is invalid")
        return field_value

    @validator("sequence")
    def validate_identifier(cls, field_value, values, field, config):
        # If sequence_type is invalid, values["sequence_type"] doesn't exist.
        field_value = field_value.upper()
        if "sequence_type" in values.keys():
            sequence_type = values["sequence_type"]
            # field_value is sequence
            target.validate_target_sequence(field_value, sequence_type)
        else:
            raise ValueError("sequence_type is invalid")
        return field_value

    @validator("label")
    def label_does_not_include_colon(cls, field_value, values, field, config) -> str:
        if isinstance(field_value, str):
            if ":" in field_value:
                raise ValidationError(f"Target sequence label `{field_value}` may not contain a colon.")

            # Sanitize the label by stripping leading/trailing whitespace and replacing any internal whitespace with
            # underscores. Fully qualified variants should never contain whitespace.
            return sanitize_target_sequence_label(field_value)

        return field_value


class TargetSequenceCreate(TargetSequenceModify):
    taxonomy: TaxonomyCreate


class TargetSequenceUpdate(TargetSequenceModify):
    pass


# Properties shared by models stored in DB
class SavedTargetSequence(TargetSequenceBase):
    record_type: str = None  # type: ignore
    taxonomy: SavedTaxonomy

    _record_type_factory = record_type_validator()(set_record_type)

    class Config:
        orm_mode = True
        arbitrary_types_allowed = True


# Properties to return to non-admin clients
class TargetSequence(SavedTargetSequence):
    taxonomy: Taxonomy


# Properties to return to admin clients
class AdminTargetSequence(SavedTargetSequence):
    creation_date: date
    modification_date: date
    taxonomy: AdminTaxonomy



from typing import Any, List

from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy import func
from sqlalchemy.orm import Session

from mavedb import deps
from mavedb.lib.identifiers import EXTERNAL_GENE_IDENTIFIER_CLASSES
from mavedb.view_models import external_gene_identifier
from mavedb.view_models.search import TextSearch

router = APIRouter(
    prefix="/api/v1/target-gene-identifiers",
    tags=["target gene identifiers"],
    responses={404: {"description": "Not found"}},
)


@router.post("/search", status_code=200, response_model=List[external_gene_identifier.ExternalGeneIdentifier])
def search_target_gene_identifiers(db_name: str, search: TextSearch, db: Session = Depends(deps.get_db)) -> Any:
    """
    Search target gene identifiers.
    """

    identifier_class = EXTERNAL_GENE_IDENTIFIER_CLASSES[db_name]
    assert hasattr(identifier_class, "identifier")

    query = db.query(identifier_class)

    if search.text and len(search.text.strip()) > 0:
        lower_search_text = search.text.strip().lower()
        query = query.filter(func.lower(identifier_class.identifier).contains(lower_search_text))
    else:
        raise HTTPException(status_code=500, detail="Search text is required")

    items = query.order_by(identifier_class.identifier).limit(50).all()
    if not items:
        items = []
    return items



import csv
import io
import logging
import re
from operator import attrgetter
from typing import Any, BinaryIO, Iterable, Optional, TYPE_CHECKING, Sequence

import numpy as np
import pandas as pd
from pandas.testing import assert_index_equal
from sqlalchemy import Integer, cast, func, or_, select
from sqlalchemy.orm import Session, aliased, contains_eager, joinedload, selectinload

from mavedb.lib.exceptions import ValidationError
from mavedb.lib.logging.context import logging_context, save_to_logging_context
from mavedb.lib.mave.constants import (
    HGVS_NT_COLUMN,
    HGVS_PRO_COLUMN,
    HGVS_SPLICE_COLUMN,
    VARIANT_COUNT_DATA,
    VARIANT_SCORE_DATA,
)
from mavedb.lib.mave.utils import is_csv_null
from mavedb.lib.validation.constants.general import null_values_list
from mavedb.lib.validation.utilities import is_null as validate_is_null
from mavedb.models.contributor import Contributor
from mavedb.models.controlled_keyword import ControlledKeyword
from mavedb.models.doi_identifier import DoiIdentifier
from mavedb.models.ensembl_identifier import EnsemblIdentifier
from mavedb.models.ensembl_offset import EnsemblOffset
from mavedb.models.experiment import Experiment
from mavedb.models.experiment_controlled_keyword import ExperimentControlledKeywordAssociation
from mavedb.models.experiment_publication_identifier import ExperimentPublicationIdentifierAssociation
from mavedb.models.experiment_set import ExperimentSet
from mavedb.models.publication_identifier import PublicationIdentifier
from mavedb.models.refseq_identifier import RefseqIdentifier
from mavedb.models.refseq_offset import RefseqOffset
from mavedb.models.score_set import ScoreSet
from mavedb.models.score_set_publication_identifier import (
    ScoreSetPublicationIdentifierAssociation,
)
from mavedb.models.target_accession import TargetAccession
from mavedb.models.target_gene import TargetGene
from mavedb.models.target_sequence import TargetSequence
from mavedb.models.taxonomy import Taxonomy
from mavedb.models.uniprot_identifier import UniprotIdentifier
from mavedb.models.uniprot_offset import UniprotOffset
from mavedb.models.user import User
from mavedb.models.variant import Variant
from mavedb.view_models.search import ScoreSetsSearch

if TYPE_CHECKING:
    from mavedb.lib.authentication import UserData
    from mavedb.lib.permissions import Action

VariantData = dict[str, Optional[dict[str, dict]]]

logger = logging.getLogger(__name__)


class HGVSColumns:
    NUCLEOTIDE: str = "hgvs_nt"  # dataset.constants.hgvs_nt_column
    TRANSCRIPT: str = "hgvs_splice"  # dataset.constants.hgvs_splice_column
    PROTEIN: str = "hgvs_pro"  # dataset.constants.hgvs_pro_column

    @classmethod
    def options(cls) -> list[str]:
        return [cls.NUCLEOTIDE, cls.TRANSCRIPT, cls.PROTEIN]


def search_score_sets(db: Session, owner_or_contributor: Optional[User], search: ScoreSetsSearch) -> list[ScoreSet]:
    save_to_logging_context({"score_set_search_criteria": search.dict()})

    query = db.query(ScoreSet)  # \
    # .filter(ScoreSet.private.is_(False))

    if owner_or_contributor is not None:
        query = query.filter(
            or_(
                ScoreSet.created_by_id == owner_or_contributor.id,
                ScoreSet.contributors.any(Contributor.orcid_id == owner_or_contributor.username),
            )
        )

    if search.published is not None:
        if search.published:
            query = query.filter(ScoreSet.published_date.isnot(None))
        else:
            query = query.filter(ScoreSet.published_date.is_(None))

    if search.text:
        lower_search_text = search.text.lower().strip()
        query = query.filter(
            or_(
                ScoreSet.urn.icontains(lower_search_text),
                ScoreSet.title.icontains(lower_search_text),
                ScoreSet.short_description.icontains(lower_search_text),
                ScoreSet.abstract_text.icontains(lower_search_text),
                ScoreSet.target_genes.any(func.lower(TargetGene.name).icontains(lower_search_text)),
                ScoreSet.target_genes.any(func.lower(TargetGene.category).icontains(lower_search_text)),
                ScoreSet.target_genes.any(
                    TargetGene.target_sequence.has(
                        TargetSequence.taxonomy.has(func.lower(Taxonomy.organism_name).icontains(lower_search_text))
                    )
                ),
                ScoreSet.target_genes.any(
                    TargetGene.target_sequence.has(
                        TargetSequence.taxonomy.has(func.lower(Taxonomy.common_name).icontains(lower_search_text))
                    )
                ),
                ScoreSet.target_genes.any(
                    TargetGene.target_accession.has(func.lower(TargetAccession.assembly).icontains(lower_search_text))
                ),
                # TODO(#94): add LICENSE, plus TAX_ID if numeric
                ScoreSet.publication_identifiers.any(
                    func.lower(PublicationIdentifier.identifier).icontains(lower_search_text)
                ),
                ScoreSet.publication_identifiers.any(
                    func.lower(PublicationIdentifier.doi).icontains(lower_search_text)
                ),
                ScoreSet.publication_identifiers.any(
                    func.lower(PublicationIdentifier.abstract).icontains(lower_search_text)
                ),
                ScoreSet.publication_identifiers.any(
                    func.lower(PublicationIdentifier.title).icontains(lower_search_text)
                ),
                ScoreSet.publication_identifiers.any(
                    func.lower(PublicationIdentifier.publication_journal).icontains(lower_search_text)
                ),
                ScoreSet.publication_identifiers.any(
                    func.jsonb_path_exists(
                        PublicationIdentifier.authors,
                        f"""$[*].name ? (@ like_regex "{lower_search_text}" flag "i")""",
                    )
                ),
                ScoreSet.doi_identifiers.any(func.lower(DoiIdentifier.identifier).icontains(lower_search_text)),
                ScoreSet.target_genes.any(
                    TargetGene.uniprot_offset.has(
                        UniprotOffset.identifier.has(
                            func.lower(UniprotIdentifier.identifier).icontains(lower_search_text)
                        )
                    )
                ),
                ScoreSet.target_genes.any(
                    TargetGene.refseq_offset.has(
                        RefseqOffset.identifier.has(
                            func.lower(RefseqIdentifier.identifier).icontains(lower_search_text)
                        )
                    )
                ),
                ScoreSet.target_genes.any(
                    TargetGene.ensembl_offset.has(
                        EnsemblOffset.identifier.has(
                            func.lower(EnsemblIdentifier.identifier).icontains(lower_search_text)
                        )
                    )
                ),
            )
        )

    if search.targets:
        query = query.filter(ScoreSet.target_genes.any(TargetGene.name.in_(search.targets)))

    if search.target_organism_names:
        query = query.filter(
            ScoreSet.target_genes.any(
                TargetGene.target_sequence.has(
                    TargetSequence.taxonomy.has(Taxonomy.organism_name.in_(search.target_organism_names))
                )
            )
        )

    if search.target_types:
        query = query.filter(ScoreSet.target_genes.any(TargetGene.category.in_(search.target_types)))

    if search.publication_identifiers:
        query = query.filter(
            ScoreSet.publication_identifiers.any(PublicationIdentifier.identifier.in_(search.publication_identifiers))
        )

    if search.databases:
        query = query.filter(ScoreSet.publication_identifiers.any(PublicationIdentifier.db_name.in_(search.databases)))

    if search.journals:
        query = query.filter(
            ScoreSet.publication_identifiers.any(PublicationIdentifier.publication_journal.in_(search.journals))
        )

    if search.authors:
        query = query.filter(
            ScoreSet.publication_identifiers.any(
                func.jsonb_path_query_array(PublicationIdentifier.authors, "$.name").op("?|")(search.authors)
            )
        )

    if search.target_accessions:
        query = query.filter(
            ScoreSet.target_genes.any(
                TargetGene.target_accession.has(TargetAccession.accession.in_(search.target_accessions))
            )
        )

    if search.keywords:
        query = query.filter(
            ScoreSet.experiment.has(
                Experiment.keyword_objs.any(
                    ExperimentControlledKeywordAssociation.controlled_keyword.has(
                        ControlledKeyword.value.in_(search.keywords)
                    )
                )
            )
        )

    score_sets: list[ScoreSet] = (
        query.join(ScoreSet.experiment)
        .options(
            contains_eager(ScoreSet.experiment).options(
                joinedload(Experiment.experiment_set),
                joinedload(Experiment.keyword_objs).joinedload(
                    ExperimentControlledKeywordAssociation.controlled_keyword
                ),
                joinedload(Experiment.created_by),
                joinedload(Experiment.modified_by),
                joinedload(Experiment.doi_identifiers),
                joinedload(Experiment.publication_identifier_associations).joinedload(
                    ExperimentPublicationIdentifierAssociation.publication
                ),
                joinedload(Experiment.raw_read_identifiers),
                selectinload(Experiment.score_sets).options(
                    joinedload(ScoreSet.doi_identifiers),
                    joinedload(ScoreSet.publication_identifier_associations).joinedload(
                        ScoreSetPublicationIdentifierAssociation.publication
                    ),
                    joinedload(ScoreSet.target_genes).options(
                        joinedload(TargetGene.ensembl_offset).joinedload(EnsemblOffset.identifier),
                        joinedload(TargetGene.refseq_offset).joinedload(RefseqOffset.identifier),
                        joinedload(TargetGene.uniprot_offset).joinedload(UniprotOffset.identifier),
                        joinedload(TargetGene.target_sequence).joinedload(TargetSequence.taxonomy),
                        joinedload(TargetGene.target_accession),
                    ),
                ),
            ),
            joinedload(ScoreSet.license),
            joinedload(ScoreSet.doi_identifiers),
            joinedload(ScoreSet.publication_identifier_associations).joinedload(
                ScoreSetPublicationIdentifierAssociation.publication
            ),
            joinedload(ScoreSet.target_genes).options(
                joinedload(TargetGene.ensembl_offset).joinedload(EnsemblOffset.identifier),
                joinedload(TargetGene.refseq_offset).joinedload(RefseqOffset.identifier),
                joinedload(TargetGene.uniprot_offset).joinedload(UniprotOffset.identifier),
                joinedload(TargetGene.target_sequence).joinedload(TargetSequence.taxonomy),
                joinedload(TargetGene.target_accession),
            ),
        )
        .order_by(Experiment.title)
        .all()
    )
    if not score_sets:
        score_sets = []

    save_to_logging_context({"matching_resources": len(score_sets)})
    logger.debug(msg=f"Score set search yielded {len(score_sets)} matching resources.", extra=logging_context())

    return score_sets  # filter_visible_score_sets(score_sets)


def fetch_superseding_score_set_in_search_result(
    score_sets: list[ScoreSet],
    requesting_user: Optional["UserData"],
    search: ScoreSetsSearch) -> list[ScoreSet]:
    """
    Remove superseded score set from search results.
    Check whether all of the score set are correct versions.
    """
    from mavedb.lib.permissions import Action
    if search.published:
        filtered_score_sets_tail = [
            find_publish_or_private_superseded_score_set_tail(
                score_set,
                Action.READ,
                requesting_user,
                search.published
            ) for score_set in score_sets
        ]
    else:
        filtered_score_sets_tail = [
            find_superseded_score_set_tail(
                score_set,
                Action.READ,
                requesting_user
            ) for score_set in score_sets
        ]
    # Remove None item.
    filtered_score_sets = [score_set for score_set in filtered_score_sets_tail if score_set is not None]
    if filtered_score_sets:
        final_score_sets = sorted(set(filtered_score_sets), key=attrgetter("urn"))
    else:
        final_score_sets = []
    return final_score_sets


def find_meta_analyses_for_experiment_sets(db: Session, urns: list[str]) -> list[ScoreSet]:
    """
    Find all score sets that are meta-analyses for score sets from a specified collection of experiment sets.

    :param db: An active database session.
    :param urns: A list of experiment set URNS.
    :return: A score set that is a meta-analysis for score sets belonging to exactly the collection of experiment sets
      specified by urns; or None if there is no such meta-analysis.
    """
    # Ensure that URNs are not repeated in the list.
    urns = list(set(urns))

    # Find all score sets that are meta-analyses for a superset of the specified URNs and are meta-analyses for
    # exactly len(urns) score sets.
    score_set_aliases = [aliased(ScoreSet) for _ in urns]
    experiment_aliases = [aliased(Experiment) for _ in urns]
    experiment_set_aliases = [aliased(ExperimentSet) for _ in urns]
    analyzed_score_set = aliased(ScoreSet)
    analyzed_experiment = aliased(Experiment)
    analyzed_experiment_set = aliased(ExperimentSet)
    urn_filters = [
        ScoreSet.meta_analyzes_score_sets.of_type(score_set_aliases[i]).any(
            score_set_aliases[i]
            .experiment.of_type(experiment_aliases[i])
            .has(
                experiment_aliases[i]
                .experiment_set.of_type(experiment_set_aliases[i])
                .has(experiment_set_aliases[i].urn == urn)
            )
        )
        for i, urn in enumerate(urns)
    ]
    return (
        db.query(ScoreSet)
        .join(ScoreSet.meta_analyzes_score_sets.of_type(analyzed_score_set))
        .join(analyzed_score_set.experiment.of_type(analyzed_experiment))
        .join(analyzed_experiment.experiment_set.of_type(analyzed_experiment_set))
        .filter(*urn_filters)
        .group_by(ScoreSet.id)
        .having(func.count(func.distinct(analyzed_experiment_set.id)) == len(urns))
        .all()
    )


def find_superseded_score_set_tail(
        score_set: ScoreSet,
        action: Optional["Action"] = None,
        user_data: Optional["UserData"] = None) -> Optional[ScoreSet]:
    from mavedb.lib.permissions import has_permission
    while score_set.superseding_score_set is not None:
        next_score_set_in_chain = score_set.superseding_score_set

        # If we were given a permission to check and the next score set in the chain does not have that permission,
        # pretend like we have reached the end of the chain. Otherwise, continue to the next score set.
        if action is not None and not has_permission(user_data, next_score_set_in_chain, action).permitted:
            return score_set

        score_set = next_score_set_in_chain

    # Handle unpublished superseding score set case.
    # The score set has a published superseded score set but has not superseding score set.
    if action is not None and not has_permission(user_data, score_set, action).permitted:
        while score_set.superseded_score_set is not None:
            next_score_set_in_chain = score_set.superseded_score_set
            if has_permission(user_data, next_score_set_in_chain, action).permitted:
                return next_score_set_in_chain
            else:
                score_set = next_score_set_in_chain
        return None

    return score_set


def find_publish_or_private_superseded_score_set_tail(
        score_set: ScoreSet,
        action: Optional["Action"] = None,
        user_data: Optional["UserData"] = None,
        publish: bool = True) -> Optional[ScoreSet]:
    from mavedb.lib.permissions import has_permission
    if publish:
        while score_set.superseding_score_set is not None:
            next_score_set_in_chain = score_set.superseding_score_set
            # Find the final published one.
            if action is not None and has_permission(user_data, score_set, action).permitted \
                    and next_score_set_in_chain.published_date is None:
                return score_set
            score_set = next_score_set_in_chain
    else:
        # Unpublished score set should not be superseded.
        # It should not have superseding score set, but possible have superseded score set.
        if action is not None and score_set.published_date is None \
                and has_permission(user_data, score_set, action).permitted:
            return score_set
        else:
            return None
    return score_set


def get_score_set_counts_as_csv(
    db: Session,
    score_set: ScoreSet,
    start: Optional[int] = None,
    limit: Optional[int] = None,
    drop_na_columns: Optional[bool] = None,
) -> str:
    assert type(score_set.dataset_columns) is dict
    count_columns = [str(x) for x in list(score_set.dataset_columns.get("count_columns", []))]
    columns = ["accession", "hgvs_nt", "hgvs_splice", "hgvs_pro"] + count_columns
    type_column = "count_data"

    variants_query = (
        select(Variant)
        .where(Variant.score_set_id == score_set.id)
        .order_by(cast(func.split_part(Variant.urn, "#", 2), Integer))
    )
    if start:
        variants_query = variants_query.offset(start)
    if limit:
        variants_query = variants_query.limit(limit)
    variants = db.scalars(variants_query).all()

    rows_data = variants_to_csv_rows(variants, columns=columns, dtype=type_column)
    if drop_na_columns:
        rows_data, columns = drop_na_columns_from_csv_file_rows(rows_data, columns)

    stream = io.StringIO()
    writer = csv.DictWriter(stream, fieldnames=columns, quoting=csv.QUOTE_MINIMAL)
    writer.writeheader()
    writer.writerows(rows_data)
    return stream.getvalue()


def get_score_set_scores_as_csv(
    db: Session,
    score_set: ScoreSet,
    start: Optional[int] = None,
    limit: Optional[int] = None,
    drop_na_columns: Optional[bool] = None,
) -> str:
    assert type(score_set.dataset_columns) is dict
    score_columns = [str(x) for x in list(score_set.dataset_columns.get("score_columns", []))]
    columns = ["accession", "hgvs_nt", "hgvs_splice", "hgvs_pro"] + score_columns
    type_column = "score_data"

    variants_query = (
        select(Variant)
        .where(Variant.score_set_id == score_set.id)
        .order_by(cast(func.split_part(Variant.urn, "#", 2), Integer))
    )
    if start:
        variants_query = variants_query.offset(start)
    if limit:
        variants_query = variants_query.limit(limit)
    variants = db.scalars(variants_query).all()

    rows_data = variants_to_csv_rows(variants, columns=columns, dtype=type_column)
    if drop_na_columns:
        rows_data, columns = drop_na_columns_from_csv_file_rows(rows_data, columns)

    stream = io.StringIO()
    writer = csv.DictWriter(stream, fieldnames=columns, quoting=csv.QUOTE_MINIMAL)
    writer.writeheader()
    writer.writerows(rows_data)
    return stream.getvalue()


def drop_na_columns_from_csv_file_rows(
    rows_data: Iterable[dict[str, Any]],
    columns: list[str]
) -> tuple[list[dict[str, Any]], list[str]]:
    """Process rows_data for downloadable CSV by removing empty columns."""
    # Convert map to list.
    rows_data = list(rows_data)
    columns_to_check = ["hgvs_nt", "hgvs_splice", "hgvs_pro"]
    columns_to_remove = []

    # Check if all values in a column are None or "NA"
    for col in columns_to_check:
        if all(validate_is_null(row[col]) for row in rows_data):
            columns_to_remove.append(col)
            for row in rows_data:
                row.pop(col, None)  # Remove column from each row

    # Remove these columns from the header list
    columns = [col for col in columns if col not in columns_to_remove]
    return rows_data, columns


null_values_re = re.compile(r"\s+|none|nan|na|undefined|n/a|null|nil", flags=re.IGNORECASE)


def is_null(value):
    """Return True if a string represents a null value."""
    value = str(value).strip().lower()
    return null_values_re.fullmatch(value) or not value


def variant_to_csv_row(variant: Variant, columns: list[str], dtype: str, na_rep="NA") -> dict[str, Any]:
    """
    Format a variant into a containing the keys specified in `columns`.

    Parameters
    ----------
    variant : variant.models.Variant
        List of variants.
    columns : list[str]
        Columns to serialize.
    dtype : str, {'scores', 'counts'}
        The type of data requested. Either the 'score_data' or 'count_data'.
    na_rep : str
        String to represent null values.

    Returns
    -------
    dict[str, Any]
    """
    row = {}
    for column_key in columns:
        if column_key == "hgvs_nt":
            value = str(variant.hgvs_nt)
        elif column_key == "hgvs_pro":
            value = str(variant.hgvs_pro)
        elif column_key == "hgvs_splice":
            value = str(variant.hgvs_splice)
        elif column_key == "accession":
            value = str(variant.urn)
        else:
            parent = variant.data.get(dtype) if variant.data else None
            value = str(parent.get(column_key)) if parent else na_rep
        if is_null(value):
            value = na_rep
        row[column_key] = value
    return row


def variants_to_csv_rows(
    variants: Sequence[Variant], columns: list[str], dtype: str, na_rep="NA"
) -> Iterable[dict[str, Any]]:
    """
    Format each variant into a dictionary row containing the keys specified in `columns`.

    Parameters
    ----------
    variants : list[variant.models.Variant]
        List of variants.
    columns : list[str]
        Columns to serialize.
    dtype : str, {'scores', 'counts'}
        The type of data requested. Either the 'score_data' or 'count_data'.
    na_rep : str
        String to represent null values.

    Returns
    -------
    list[dict[str, Any]]
    """
    return map(lambda v: variant_to_csv_row(v, columns, dtype, na_rep), variants)


def find_meta_analyses_for_score_sets(db: Session, urns: list[str]) -> list[ScoreSet]:
    """
    Find all score sets that are meta-analyses for a specified collection of other score sets.

    :param db: An active database session.
    :param urns: A list of score set URNS.
    :return: A score set that is a meta-analysis for exactly the collection of score sets specified by urns; or None if
      there is no such meta-analysis.
    """
    # Ensure that URNs are not repeated in the list.
    urns = list(set(urns))

    # Find all score sets that are meta-analyses for a superset of the specified URNs and are meta-analyses for
    # exactly len(urns) score sets.
    score_set_aliases = [aliased(ScoreSet) for _ in urns]
    analyzed_score_set = aliased(ScoreSet)
    urn_filters = [
        ScoreSet.meta_analyzes_score_sets.of_type(score_set_aliases[i]).any(score_set_aliases[i].urn == urn)
        for i, urn in enumerate(urns)
    ]
    return (
        db.query(ScoreSet)
        .join(ScoreSet.meta_analyzes_score_sets.of_type(analyzed_score_set))
        .filter(*urn_filters)
        .group_by(ScoreSet.id)
        .having(func.count(analyzed_score_set.id) == len(urns))
        .all()
    )


def filter_visible_score_sets(items: list[ScoreSet]):
    # TODO Take the user into account.
    return filter(lambda item: not item.private, items or [])


def arrays_equal(array1: np.ndarray, array2: np.ndarray):
    # if the shape isn't the same the arrays are different.
    # otherwise for each value make sure either both values are null
    # or the values are equal.
    return array1.shape == array2.shape and all(
        # note that each of the three expressions here is a boolean ndarray
        # so combining them with bitwise `&` and `|` works:
        (pd.isnull(array1) & pd.isnull(array2)) | (array1 == array2)
    )


def validate_datasets_define_same_variants(scores, counts):
    """
    Checks if two `pd.DataFrame` objects parsed from uploaded files
    define the same variants.

    Parameters
    ----------
    scores : `pd.DataFrame`
        Scores dataframe parsed from an uploaded scores file.
    counts : `pd.DataFrame`
        Scores dataframe parsed from an uploaded counts file.
    """
    # TODO First, confirm that the two dataframes have the same HGVS columns.
    if any(
        col in scores and not arrays_equal(scores[col].sort_values(), counts[col].sort_values())
        for col in (HGVS_NT_COLUMN, HGVS_SPLICE_COLUMN, HGVS_PRO_COLUMN)
    ):
        raise ValidationError(
            "Your score and counts files do not define the same variants. "
            "Check that the hgvs columns in both files match."
        )


def create_variants_data(scores, counts=None, index_col=None) -> list[VariantData]:
    """
    Given two `defaultdict`s `score_map` and `count_map`, create an
    `OrderedDict` indexed by `hgvs_nt` where the keys are the attribute
    fields required to instantiate a `variant.models.Variant` instance.

    NOTE: Assumes that the dataframes are indexed by their primary columns,
    and that they define the same variants in both hgvs columns.

    Parameters
    ----------
    scores : Union[`pd.DataFrame`, str]
        Map indexed by the primary hgvs column inferred during validation.
        Map values are `dict` records where the key-pairs are column-value
        pairs inferred from the `scores` file uploaded during submission.
    counts : Union[`pd.DataFrame`, str] optional
        Map indexed by the primary hgvs column inferred during validation.
        Map values are `dict` records where the key-pairs are column-value
        pairs inferred from the `counts` file uploaded during submission.
    index_col : str
        Column to use as index, which is used when grouping rows between
        dataframes.

    Returns
    -------
    `list`
        Formatted records that can be used to create `variant.models.Variant`
        instances.
    """

    if isinstance(scores, str):
        scores = pd.read_json(scores, orient="records")
    if isinstance(counts, str):
        counts = pd.read_json(counts, orient="records")

    has_count_data = counts is not None and len(counts) > 0
    has_score_data = scores is not None and len(scores) > 0

    if index_col:
        scores.index = pd.Index(scores[index_col])
        if has_count_data:
            counts.index = pd.Index(counts[index_col])

    if not has_score_data:
        return []

    if has_count_data:
        assert_index_equal(scores.index.sort_values(), counts.index.sort_values())
        validate_datasets_define_same_variants(scores, counts)

    variants = []
    for primary_hgvs, group in scores.groupby(by=scores.index, sort=False):
        score_records = group.to_dict(orient="records")
        if has_count_data:
            count_records = counts[counts.index == primary_hgvs].to_dict(orient="records")
            assert len(score_records) == len(count_records)
        else:
            # Make duplicates to zip with self when no count data.
            count_records = [r.copy() for r in score_records]

        for sr, cr in zip(score_records, count_records):
            hgvs_nt = sr.pop(HGVS_NT_COLUMN)
            hgvs_splice = sr.pop(HGVS_SPLICE_COLUMN)
            hgvs_pro = sr.pop(HGVS_PRO_COLUMN)

            if is_csv_null(hgvs_nt) or hgvs_nt is np.NaN or hgvs_nt == "nan":
                hgvs_nt = None
            if is_csv_null(hgvs_splice) or hgvs_splice is np.NaN or hgvs_splice == "nan":
                hgvs_splice = None
            if is_csv_null(hgvs_pro) or hgvs_pro is np.NaN or hgvs_pro == "nan":
                hgvs_pro = None

            cr.pop(HGVS_NT_COLUMN)
            cr.pop(HGVS_SPLICE_COLUMN)
            cr.pop(HGVS_PRO_COLUMN)

            # JSON fields cannot store np.NaN values, so convert np.NaN to None.
            for key, value in sr.items():
                if is_csv_null(value) or value is np.NaN:
                    sr[key] = None
            if cr:
                for key, value in cr.items():
                    if is_csv_null(value) or value is np.NaN:
                        cr[key] = None

            data = {VARIANT_SCORE_DATA: sr, VARIANT_COUNT_DATA: {} if cr == sr else cr}
            variant = {
                HGVS_NT_COLUMN: hgvs_nt,
                HGVS_SPLICE_COLUMN: hgvs_splice,
                HGVS_PRO_COLUMN: hgvs_pro,
                "data": data,
            }
            variants.append(variant)

    return variants


def create_variants(db, score_set: ScoreSet, variants_data: list[VariantData], batch_size=None) -> int:
    num_variants = len(variants_data)
    variant_urns = bulk_create_urns(num_variants, score_set, True)
    variants = (
        # TODO: Is there a nicer way to handle this than passing dicts into kwargs
        # of the class initializer?
        Variant(urn=urn, score_set_id=score_set.id, **kwargs)  # type: ignore
        for urn, kwargs in zip(variant_urns, variants_data)
    )
    db.bulk_save_objects(variants)
    db.add(score_set)
    return len(score_set.variants)


def refresh_variant_urns(db: Session, score_set: ScoreSet):
    variants = db.execute(select(Variant).where(Variant.score_set_id == score_set.id)).scalars()

    for variant in variants:
        if not variant.urn:
            raise ValueError("All variants should have an associated URN.")

        variant_number = variant.urn.split("#")[1]
        refreshed_urn = f"{score_set.urn}#{variant_number}"
        variant.urn = refreshed_urn
        db.add(variant)

    db.commit()


def bulk_create_urns(n, score_set, reset_counter=False) -> list[str]:
    start_value = 0 if reset_counter else score_set.num_variants
    parent_urn = score_set.urn
    child_urns = ["{}#{}".format(parent_urn, start_value + (i + 1)) for i in range(n)]
    current_value = start_value + n
    score_set.num_variants = current_value
    return child_urns


def csv_data_to_df(file_data: BinaryIO) -> pd.DataFrame:
    extra_na_values = list(
        set(
            list(null_values_list)
            + [str(x).lower() for x in null_values_list]
            + [str(x).upper() for x in null_values_list]
            + [str(x).capitalize() for x in null_values_list]
        )
    )

    ingested_df = pd.read_csv(
        filepath_or_buffer=file_data,
        sep=",",
        encoding="utf-8",
        quotechar="'",
        index_col=False,
        na_values=extra_na_values,
        keep_default_na=True,
        dtype={**{col: str for col in HGVSColumns.options()}, "scores": float},
    )

    for c in HGVSColumns.options():
        if c not in ingested_df.columns:
            ingested_df[c] = np.NaN

    return ingested_df


def columns_for_dataset(dataset: Optional[pd.DataFrame]) -> list[str]:
    if dataset is None:
        return []

    return [col for col in dataset.columns if col not in HGVSColumns.options()]



import os
from typing import Mapping, Optional, Union

import eutils  # type: ignore
from eutils import EutilsNCBIError  # type: ignore
from eutils._internal.xmlfacades.pubmedarticle import PubmedArticle  # type: ignore
from eutils._internal.xmlfacades.pubmedarticleset import PubmedArticleSet  # type: ignore
from idutils import is_doi, normalize_doi
from sqlalchemy import select
from sqlalchemy.orm import Session

from mavedb.lib.exceptions import AmbiguousIdentifierError, NonexistentIdentifierError
from mavedb.lib.external_publications import Crossref, CrossrefWork, PublicationAuthors, Rxiv, RxivContentDetail
from mavedb.lib.validation.publication import identifier_valid_for, infer_identifier_from_url, validate_db_name
from mavedb.models.doi_identifier import DoiIdentifier
from mavedb.models.ensembl_identifier import EnsemblIdentifier
from mavedb.models.ensembl_offset import EnsemblOffset
from mavedb.models.publication_identifier import PublicationIdentifier
from mavedb.models.raw_read_identifier import RawReadIdentifier
from mavedb.models.refseq_identifier import RefseqIdentifier
from mavedb.models.refseq_offset import RefseqOffset
from mavedb.models.target_gene import TargetGene
from mavedb.models.uniprot_identifier import UniprotIdentifier
from mavedb.models.uniprot_offset import UniprotOffset

# XXX these classes all have an "identifier" attribute but there's no superclass
# to unify them ...

EXTERNAL_GENE_IDENTIFIER_CLASSES = {
    "Ensembl": EnsemblIdentifier,
    "RefSeq": RefseqIdentifier,
    "UniProt": UniprotIdentifier,
}

EXTERNAL_GENE_IDENTIFIER_OFFSET_CLASSES = {"Ensembl": EnsemblOffset, "RefSeq": RefseqOffset, "UniProt": UniprotOffset}

EXTERNAL_GENE_IDENTIFIER_OFFSET_ATTRIBUTES = {
    "Ensembl": "ensembl_offset",
    "RefSeq": "refseq_offset",
    "UniProt": "uniprot_offset",
}


class ExternalPublication:
    """
    Class for a generic External Publication object, which may be a response
    from the public APIs of any accepted publication db.
    """

    identifier: str
    db_name: str

    title: str
    authors: list[PublicationAuthors]

    abstract: Optional[str]
    doi: Optional[str]
    publication_year: Optional[int]
    publication_volume: Optional[str]
    publication_pages: Optional[str]
    publication_journal: Optional[str]

    supplied_url: Optional[str] = None

    _article_cit_fmt = "{author}. {title}. {journal}. {year}; {volume}:{pages}. {doi}"

    def __init__(
        self,
        identifier: str,
        db_name: str,
        external_publication: Union[RxivContentDetail, PubmedArticle, CrossrefWork],
    ) -> None:
        validate_db_name(db_name)

        # Required identification fields
        self.identifier = identifier
        self.db_name = db_name

        # Shared fields
        self.title = str(external_publication.title)
        self.abstract = str(external_publication.abstract) if external_publication.abstract else None
        self.doi = str(external_publication.doi) if external_publication.doi else None

        # Non-shared fields
        if isinstance(external_publication, PubmedArticle):
            self.authors = self._infer_author_list(external_publication.authors)
            self.publication_year = int(external_publication.year) if external_publication.year else None
            self.publication_journal = str(external_publication.jrnl) if external_publication.jrnl else None
            self.publication_volume = str(external_publication.volume) if external_publication.volume else None
            self.publication_pages = str(external_publication.pages) if external_publication.pages else None

        elif isinstance(external_publication, RxivContentDetail):
            self.authors = external_publication.authors
            self.publication_year = external_publication.date.year if external_publication.date else None
            self.publication_journal = "Preprint"  # blanket `Preprint` journal for preprint articles
            self.publication_volume = None
            self.publication_pages = None

        elif isinstance(external_publication, CrossrefWork):
            self.authors = external_publication.authors
            self.supplied_url = external_publication.url
            self.publication_year = external_publication.publication_year
            self.publication_journal = external_publication.publication_journal
            self.publication_volume = external_publication.volume
            self.publication_pages = None

    def _infer_author_list(self, authors: list[str]) -> list[PublicationAuthors]:
        """
        Generates a list of author names and thier authorship level associated with this publication.
        """
        return [{"name": author, "primary": idx == 0} for idx, author in enumerate(authors)]

    def _format_authors(self) -> str:
        """Helper function for returning a well formatted HTML author list"""
        if self.authors and len(self.authors) > 2:
            author = str(self.authors[0]["name"]) + ", <i>et al</i>"
        elif self.authors and len(self.authors) == 2:
            author = " and ".join([str(author["name"]) for author in self.authors])
        elif self.authors and len(self.authors) < 2:
            author = str(self.authors[0]["name"])
        else:
            author = ""

        return author

    @property
    def primary_author(self) -> str:
        (primary_author,) = (author["name"] for author in self.authors if author["primary"])
        return primary_author

    @property
    def secondary_authors(self) -> list[str]:
        return [author["name"] for author in self.authors if not author["primary"]]

    @property
    def url(self) -> Optional[str]:
        if self.db_name == "PubMed":
            return f"http://www.ncbi.nlm.nih.gov/pubmed/{self.identifier}"
        elif self.db_name == "bioRxiv":
            return f"https://www.biorxiv.org/content/10.1101/{self.identifier}"
        elif self.db_name == "medRxiv":
            return f"https://www.medrxiv.org/content/10.1101/{self.identifier}"
        elif self.db_name == "Crossref":
            return self.supplied_url
        else:
            return None

    @property
    def reference_html(self) -> str:
        """
        Return a well formatted citation HTML string based on article data.
        Intends to return an identical citation html string to metapub.PubMedArticle.
        """
        author = self._format_authors()

        doi_str = "" if not self.doi else self.doi
        title = "(None)" if not self.title else self.title.strip(".")
        journal = (
            "(None)"
            if not self.publication_journal or self.publication_journal == "Preprint"
            else self.publication_journal.strip(".")
        )
        year = "(Unknown year)" if not self.publication_year else self.publication_year
        volume = "(Unknown volume)" if not self.publication_volume else self.publication_volume
        pages = "(Unknown pages)" if not self.publication_pages else self.publication_pages

        return self._article_cit_fmt.format(
            author=author, volume=volume, pages=pages, year=year, title=title, journal=journal, doi=doi_str
        )


async def find_or_create_doi_identifier(db: Session, identifier: str):
    """
    Find an existing DOI identifier record with the specified identifier string, or create a new one.

    :param db: An active database session
    :param identifier: A valid DOI identifier
    :return: An existing DoiIdentifier containing the specified identifier string, or a new, unsaved DoiIdentifier
    """
    doi_identifier = db.query(DoiIdentifier).filter(DoiIdentifier.identifier == identifier).one_or_none()
    if not doi_identifier:
        doi_identifier = DoiIdentifier(identifier=identifier, db_name="DOI", url=f"https://doi.org/{identifier}")
    return doi_identifier


async def fetch_pubmed_article(identifier: str) -> Optional[PubmedArticle]:
    """
    Fetch an existing PubMed article from NCBI
    """
    fetch = eutils.QueryService(api_key=os.getenv("NCBI_API_KEY"))
    try:
        fetched_articles = list(PubmedArticleSet(fetch.efetch({"db": "pubmed", "id": identifier})))
        assert len(fetched_articles) < 2

    except AssertionError as exc:
        raise AmbiguousIdentifierError(f"Fetched more than 1 PubMed article associated with PMID {identifier}") from exc
    except EutilsNCBIError:
        return None

    if fetched_articles:
        return fetched_articles[0]
    else:
        return None


# TODO: Could search on article_detail -> content_detail to try and get the richer
#       metadata that exists on already published articles. leaving for now since
#       the main purpose of these changes is to support preprints, which we now do,
#       and adding any additional API calls would slow this fetch process down.
#       - capodb 2023.06.06
#
# NOTE: The most up to date version of a preprint will be the last element in the
#       content detail list.
async def fetch_biorxiv_article(identifier: str) -> Optional[RxivContentDetail]:
    """
    Fetch an existing bioRxiv article from Rxiv
    """
    fetch = Rxiv("https://api.biorxiv.org", "biorxiv")
    articles = fetch.content_detail(identifier=identifier)
    try:
        return articles[-1]
    except IndexError:
        return None


async def fetch_medrxiv_article(identifier: str) -> Optional[RxivContentDetail]:
    """
    Fetch an existing medRxiv article from Rxiv
    """
    fetch = Rxiv("https://api.biorxiv.org", "medrxiv")
    articles = fetch.content_detail(identifier=identifier)
    try:
        return articles[-1]
    except IndexError:
        return None


async def fetch_crossref_work(identifier: str) -> Optional[CrossrefWork]:
    fetch = Crossref(endpoint="works")
    return fetch.doi(identifier)


async def find_generic_article(
    db: Session, identifier: str, db_name: Optional[str] = None
) -> Mapping[str, Union[PublicationIdentifier, ExternalPublication, None]]:
    """
    Check if a provided publication identifier ambiguously identifies a publication,
    ie the same identifier is identifies publications in multiple publication databases
    that we accept.

    :param db: An active database session
    :param identifier: A valid publication identifier
    :return: A list of databases where this identifier exists.
    """
    db_specific_fetches = {
        "Crossref": fetch_crossref_work,
        "PubMed": fetch_pubmed_article,
        "bioRxiv": fetch_biorxiv_article,
        "medRxiv": fetch_medrxiv_article,
    }

    # We also accept URLs from our accepted publications. Attempt to convert a potential URL to an identifier.
    identifier = infer_identifier_from_url(identifier)

    # Only check entries with the appropriate `db_name` if one is provided.
    db_specific_match: dict[str, Union[PublicationIdentifier, ExternalPublication, None]] = {}
    if db_name:
        if db_name == "Crossref":
            internal_publication_query = select(PublicationIdentifier).filter(PublicationIdentifier.doi == identifier)
        else:
            internal_publication_query = select(PublicationIdentifier).filter(
                PublicationIdentifier.identifier == identifier
            )

        existing_publication = db.execute(
            internal_publication_query.filter(PublicationIdentifier.db_name == db_name)
        ).scalar_one_or_none()

        if not existing_publication:
            external_publication = await db_specific_fetches[db_name](identifier)
            db_specific_match = {
                db_name: (
                    ExternalPublication(identifier, db_name, external_publication) if external_publication else None
                )
            }
        else:
            db_specific_match = {db_name: existing_publication}

        return db_specific_match

    # If the identifier is a DOI, it will necessarily have a unique match in Crossref (if such a match exists).
    # Return this match directly where possible.
    found_articles: dict[str, Union[PublicationIdentifier, ExternalPublication, None]] = {}
    if is_doi(identifier):
        identifier = normalize_doi(identifier)

        existing_publication = db.execute(
            select(PublicationIdentifier).filter(PublicationIdentifier.doi == identifier)
        ).scalar_one_or_none()

        if not existing_publication:
            external_publication = await fetch_crossref_work(identifier)
            found_articles["Crossref"] = (
                ExternalPublication(identifier, "Crossref", external_publication) if external_publication else None
            )
        else:
            # When we find an existing publication via DOI, it is not always the case that it came from Crossref originally.
            # Use the existing publication db as the article key if it exists, otherwise default to Crossref since this is a DOI.
            existing_db_name = existing_publication.db_name if existing_publication.db_name else "Crossref"
            found_articles[existing_db_name] = existing_publication

        return found_articles

    # When we are not provided a db name, we must try to match the provided identifier to a
    # publication from each one of our accepted databases.
    for publication_db, identifier_valid in identifier_valid_for(identifier).items():
        if identifier_valid:
            existing_publication = db.execute(
                select(PublicationIdentifier)
                .filter(PublicationIdentifier.identifier == identifier)
                .filter(PublicationIdentifier.db_name == publication_db)
            ).scalar_one_or_none()

            if not existing_publication:
                external_publication = await db_specific_fetches[publication_db](identifier)
                found_articles[publication_db] = (
                    ExternalPublication(identifier, publication_db, external_publication)
                    if external_publication
                    else None
                )
            else:
                found_articles[publication_db] = existing_publication

    return found_articles


def create_generic_article(article: ExternalPublication) -> PublicationIdentifier:
    """
    Create a new (unsaved) publication identifier object based on the provided identifier, article metadata,
    and publication database name.
    """
    return PublicationIdentifier(
        identifier=article.identifier,
        db_name=article.db_name,
        url=article.url,
        title=article.title,
        abstract=article.abstract,
        authors=article.authors,
        publication_year=article.publication_year,
        doi=article.doi,
        publication_journal=article.publication_journal,
        reference_html=article.reference_html,
    )


async def find_or_create_publication_identifier(
    db: Session, identifier: str, db_name: Optional[str] = None
) -> PublicationIdentifier:
    """
    Find an existing publication identifier record with the specified identifier string, or create a new one.

    :param db: An active database session
    :param identifier: A valid publication identifier
    :return: An existing PublicationIdentifier containing the specified identifier string, or a new, unsaved PublicationIdentifier
    """
    article_matches = await find_generic_article(db, identifier, db_name)

    if not any(article_matches.values()):
        raise NonexistentIdentifierError(
            f"No matching articles found for identifier {identifier} across all accepted publication databases."
        )

    if sum(article is not None for article in article_matches.values()) > 1:
        raise AmbiguousIdentifierError(
            f"Found multiple articles associated with identifier {identifier}. Specify a `db_name` along with this identifier to avoid ambiguity."
        )

    matched_article = next(article for article in article_matches.values() if article is not None)

    # If the article already exists, return it directly.
    if isinstance(matched_article, PublicationIdentifier):
        return matched_article

    # TODO(#214): It may be useful for internal consistency to use the Crossref record fetched via DOI if it exists. If a publication did not
    #             have a DOI, we would need to use the record as returned by PubMed/bioRxiv/medRxiv.

    return create_generic_article(matched_article)


async def find_or_create_raw_read_identifier(db: Session, identifier: str):
    raw_read_identifier = db.query(RawReadIdentifier).filter(RawReadIdentifier.identifier == identifier).one_or_none()
    if not raw_read_identifier:
        raw_read_identifier = RawReadIdentifier(
            identifier=identifier, db_name="SRA", url=f"http://www.ebi.ac.uk/ena/data/view/{identifier}"
        )
    return raw_read_identifier


async def find_or_create_external_gene_identifier(db: Session, db_name: str, identifier: str):
    """
    Find an existing gene identifier record with the specified gene database name and identifier string, or create a new
    one.

    :param db: An active database session
    :param identifier: A valid identifier
    :return: An existing EnsemblIdentifier, RefseqIdentifier, or UniprotIdentifier containing the specified identifier
      string, or a new, unsaved instance of one of these classes
    """

    # TODO Handle key errors.
    identifier_class = EXTERNAL_GENE_IDENTIFIER_CLASSES[db_name]
    assert hasattr(identifier_class, "identifier")

    external_gene_identifier = (
        db.query(identifier_class).filter(identifier_class.identifier == identifier).one_or_none()
    )

    if not external_gene_identifier:
        external_gene_identifier = identifier_class(
            identifier=identifier,
            db_name=db_name,
            # TODO Set URL from identifier
            # url=f'https://doi.org/{identifier}'
        )
    return external_gene_identifier


async def create_external_gene_identifier_offset(
    db: Session, target_gene: TargetGene, db_name: str, identifier: str, offset: int
):
    external_gene_identifier = await find_or_create_external_gene_identifier(db, db_name, identifier)

    # TODO Handle key errors.
    offset_class = EXTERNAL_GENE_IDENTIFIER_OFFSET_CLASSES[external_gene_identifier.db_name]
    external_gene_identifier_offset = offset_class(identifier=external_gene_identifier, offset=offset)

    identifier_offset_attribute = EXTERNAL_GENE_IDENTIFIER_OFFSET_ATTRIBUTES[external_gene_identifier.db_name]
    setattr(target_gene, identifier_offset_attribute, external_gene_identifier_offset)



"""
(Extremely) Minimal library for interacting with Rxiv and Crossref APIs
"""

import datetime
import json
import logging
import re
from typing import Any, Optional, TypedDict, Union

import requests
from idutils import is_doi

logger = logging.getLogger(__name__)


class PublicationAuthors(TypedDict):
    name: str
    primary: bool


class CrossrefObject:
    doi: str
    identifier: str

    def __init__(self, identifier: str) -> None:
        assert is_doi(identifier)
        self.doi = self.identifier = identifier


class CrossrefWork(CrossrefObject):
    title: str
    abstract: Optional[str]
    url: Optional[str]
    authors: list[PublicationAuthors]
    publication_year: Optional[int]
    volume: Optional[str]
    publication_journal: Optional[str]

    def __init__(self, resource: dict[str, Any]) -> None:
        super().__init__(resource["DOI"])

        # title is within a list.
        self.title = resource["title"][0]
        self.url = resource.get("URL")
        self.volume = resource.get("volume")

        # Strip HTML tags from returned abstract text. Strip pre-pended `Abstract` from resulting string.
        self.abstract = resource.get("abstract")
        if self.abstract:
            self.abstract = re.compile(r"(<!--.*?-->|<[^>]*>)").sub("", self.abstract).strip("Abstract")

        # Publication journal is contained within a list field.
        container_title: Optional[list[str]] = resource.get("container-title")
        if container_title:
            # Empty lists are falsy, so we are safe from IndexErrors here.
            self.publication_journal = container_title[0]
        else:
            self.publication_journal = None

        # Publication date is contained within a list of date parts.
        publication_date: Optional[dict[str, list[list[int]]]] = resource.get("published")
        if publication_date:
            try:
                # publication_date = {"date-parts": [[Y, M, D]]}
                self.publication_year = publication_date["date-parts"][0][0]
            # Some publications may only have year and month information.
            except IndexError:
                self.publication_year = None
        else:
            self.publication_year = None

        # Construct internally styled author list from the resources' list of authors.
        authors: list[dict[str, str]] = resource.get("author", [])
        self.authors = [
            {
                "name": f"{author.get('given', '').strip()} {author.get('family', '').strip()}",
                "primary": True if author.get("sequence") == "first" else False,
            }
            for author in authors
        ]


class CrossrefAgency(CrossrefObject):
    id: Optional[str]
    label: Optional[str]

    def __init__(self, resource: dict[str, Any]) -> None:
        super().__init__(resource["DOI"])

        agency: dict = resource.get("agency", {})
        self.id = agency.get("id")
        self.label = agency.get("label")


class RxivPublication:
    """
    Class for generic returned Rxiv publication metadata
    """

    title: str
    preprint_doi: Optional[str]
    published_doi: Optional[str]
    category: Optional[str]
    preprint_date: Optional[datetime.date]
    published_date: Optional[datetime.date]

    def __init__(self, metadata: dict[str, str]) -> None:
        self.title = metadata.get("title", metadata.get("preprint_title", ""))
        self.preprint_doi = metadata.get("preprint_doi")
        self.published_doi = metadata.get("published_doi")
        self.preprint_category = metadata.get("preprint_category")

        preprint_date = metadata.get("preprint_date")
        published_date = metadata.get("published_date")
        self.preprint_date = datetime.datetime.strptime(preprint_date, "%Y-%m-%d") if preprint_date else None
        self.published_date = datetime.datetime.strptime(published_date, "%Y-%m-%d") if published_date else None

    def generate_author_list(self, metadata: dict[str, str]) -> list[PublicationAuthors]:
        authors = [s.strip() for s in metadata.get("preprint_authors", "").split(";")]
        return [{"name": author, "primary": idx == 0} for idx, author in enumerate(authors)]


class RxivContentDetail(RxivPublication):
    """
    Class for generic returned Rxiv publication content (non-published)
    """

    authors: list[PublicationAuthors]

    doi: Optional[str]
    author_corresponding: Optional[str]
    author_corresponding_institution: Optional[str]
    date: Optional[datetime.date]
    version: Optional[str]
    type: Optional[str]
    license: Optional[str]
    jatsxml: Optional[str]
    abstract: Optional[str]
    published: Optional[str]
    server: Optional[str]

    def __init__(self, metadata: dict[str, str]) -> None:
        super().__init__(metadata)

        if metadata.get("doi"):
            self.doi = metadata["doi"]
        else:
            self.doi = self.published_doi if self.published_doi else self.preprint_doi

        self.author_corresponding = metadata.get("author_corresponding")
        self.author_corresponding_institution = metadata.get("author_corresponding_institution")
        self.version = metadata.get("version")
        self.type = metadata.get("type")
        self.license = metadata.get("license")
        self.jatsxml = metadata.get("jatsxml")
        self.abstract = metadata.get("abstract")
        self.published = metadata.get("published")
        self.server = metadata.get("server")

        if metadata.get("date"):
            self.date = datetime.datetime.strptime(metadata["date"], "%Y-%m-%d")
        else:
            self.date = self.published_date if self.published_date else self.preprint_date

        self.authors = self.generate_author_list(metadata)


class RxivPublisherDetail(RxivPublication):
    """
    Class for generic returned Rxiv publisher metadata
    """

    published_citation_count: Optional[str]

    def __init__(self, metadata: dict[str, str]) -> None:
        super().__init__(metadata)
        self.published_citation_count = metadata.get("published_citation_count")


class RxivPublicationDetail(RxivPublication):
    """
    Class for generic returned Rxiv article metadata (published)
    """

    authors: list[PublicationAuthors]
    author_corresponding: Optional[str]
    author_corresponding_institution: Optional[str]
    platform: Optional[str]
    abstract: Optional[str]
    published_journal: Optional[str]

    _article_cit_fmt = "{author}. {title}. {journal}. {year}; {volume}:{pages}.{doi}"

    def __init__(self, metadata: dict[str, str]) -> None:
        """
        NOTE: We assume here that the first author in each of these author lists is the primary author
              of a publication. From what I have seen so far from the metapub and biorxiv APIs, this
              is a fine assumption to make, but it doesn't come with future guarantees and this may
              not be the case for certain publications.
        """
        super().__init__(metadata)
        self.author_corresponding = metadata.get("preprint_author_corresponding")
        self.author_corresponding_institution = metadata.get("preprint_author_corresponding_institution")
        self.platform = metadata.get("preprint_platform")
        self.abstract = metadata.get("preprint_abstract")
        self.published_journal = metadata.get("published_journal")

        self.authors = self.generate_author_list(metadata)


class RxivStatistics:
    interval: str

    def __init__(self, metadata: dict[str, Union[str, int]]) -> None:
        self.interval = str(metadata.get("month", metadata["year"]))  # interval will be one of these fields


class RxivContentStatistics(RxivStatistics):
    new_papers: int
    new_papers_cumulative: int
    revised_papers: int
    revised_papers_cumulative: int

    def __init__(self, metadata: dict[str, Union[str, int]]) -> None:
        super().__init__(metadata)
        self.new_papers = int(metadata["new_papers"])
        self.new_papers_cumulative = int(metadata["new_papers_cumulative"])
        self.revised_papers = int(metadata["revised_papers_cumulative"])
        self.revised_papers_cumulative = int(metadata["revised_papers_cumulative"])


class RxivUsageStatistics(RxivStatistics):
    abstract_views: int
    full_text_views: int
    pdf_downloads: int
    abstract_cumulative: int
    full_text_cumulative: int
    pdf_cumulative: int

    def __init__(self, metadata: dict[str, Union[str, int]]) -> None:
        super().__init__(metadata)
        self.abstract_views = int(metadata["abstract_views"])
        self.full_text_views = int(metadata["full_text_views"])
        self.pdf_downloads = int(metadata["pdf_downloads"])
        self.abstract_cumulative = int(metadata["abstract_cumulative"])
        self.full_text_cumulative = int(metadata["full_text_cumulative"])
        self.pdf_cumulative = int(metadata["pdf_cumulative"])


class Crossref:
    url = "https://api.crossref.org"
    endpoint: str

    def __init__(self, endpoint) -> None:
        super().__init__()
        self.endpoint = endpoint

    def _fetch(self, url) -> Optional[dict]:
        result = requests.get(url)

        if result.status_code == 404:
            return None

        result.raise_for_status()
        return result.json()["message"]

    def doi(self, identifier: str) -> Optional[CrossrefWork]:
        result = self._fetch(f"{self.url}/{self.endpoint}/{identifier}")
        return CrossrefWork(result) if result else None

    def agency(self, identifier: str) -> Optional[CrossrefAgency]:
        result = self._fetch(f"{self.url}/{self.endpoint}/{identifier}/agency")
        return CrossrefAgency(result) if result else None


class Rxiv:
    """
    Interact with the bioRxiv and medRxiv pre-print servers via thier web APIs.

    For more detailed API documentation, see:
    - https://api.biorxiv.org/
    - https://api.medrxiv.org/
    """

    api_url: str
    server: Optional[str]

    def __init__(self, api_url: str, server: Optional[str]) -> None:
        self.api_url = api_url
        self.server = server

    def content_detail(
        self,
        server: Optional[str] = None,
        identifier: Optional[str] = None,
        interval: Optional[tuple[str, str]] = None,
        cursor: int = 0,
        format: str = "json",
    ) -> list[RxivContentDetail]:
        """
        Fetch content detail for an article matching the passed `identifier` or for
        articles that match the provided `interval`. This content has generally
        not yet been published.

        The optional `server` parameter can be used to specify the pre-print
        server on which to search on. If this parameter is not provided, the
        server given in `self.server` will be used. One of these values must
        be set.

        The optional `identifier` parameter can be used to filter content that matches
        the provided identifier.

        The optional `interval` parameter can be used to filter content so that
        it matches the provided interval. Valid intervals are:
        - A tuple containing two YYYY-MM-DD dates. Note that the earlier of the
          two dates should appear first.

        The `cursor` parameter can be used to move the start cursor for the search.
        For instance, passing the number '5' will show matching results starting
        from the 5th piece of content. Default: 0

        The `format` parameter can be used to specify the return format of this
        method. Accepted values are `json` and `xml`.

        NOTE: This API endpoint only supports the searching via either `interval`
              or `identifier`, not both.
        """
        search_server = server if server else self.server
        if not search_server:
            raise ValueError("No server provided with default set as `None`.")

        if (interval) and (identifier):
            raise ValueError(
                "Searching by `interval` or `cursor` and `identifier` is not supported. Please provide one or the other."
            )

        if (not interval) and (not identifier):
            raise ValueError("To search for article details, you must supply either an `identifier` or an `interval`.")

        if identifier:
            return [
                RxivContentDetail(article_detail)
                for article_detail in self._fetch_by_identifier(
                    server=search_server, content="details", identifier=identifier, format=format
                ).get("collection", [])
            ]

        if interval:
            return [
                RxivContentDetail(article_detail)
                for article_detail in self._fetch_by_interval(
                    server=search_server, content="details", interval=interval, cursor=cursor, format=format
                ).get("collection", [])
            ]

        return []

    def article_detail(
        self,
        server: Optional[str] = None,
        identifier: Optional[str] = None,
        interval: Optional[tuple[str, str]] = None,
        cursor: int = 0,
        format: str = "json",
    ) -> list[RxivPublicationDetail]:
        """
        Fetch article detail for an article matching the passed `identifier` or for
        articles that match the provided `interval`. This content has generally
        already been published.

        The optional `server` parameter can be used to specify the pre-print
        server on which to search on. If this parameter is not provided, the
        server given in `self.server` will be used. One of these values must
        be set.

        The optional `identifier` parameter can be used to filter content that matches
        the provided identifier.

        The optional `interval` parameter can be used to filter content so that
        it matches the provided interval. Valid intervals are:
        - A tuple containing two YYYY-MM-DD dates. Note that the earlier of the
          two dates should appear first.

        The `cursor` parameter can be used to move the start cursor for the search.
        For instance, passing the number '5' will show matching results starting
        from the 5th piece of content. Default: 0

        The `format` parameter can be used to specify the return format of this
        method. Accepted values are `json` and `xml`.

        NOTE: This API endpoint only supports the searching via either `interval`
              or `identifier`, not both.
        """
        search_server = server if server else self.server
        if not search_server:
            raise ValueError("No server provided with default set as `None`.")

        if (interval) and (identifier):
            raise ValueError(
                "Searching by `interval` or `cursor` and `identifier` is not supported. Please provide one or the other."
            )

        if (not interval) and (not identifier):
            raise ValueError("To search for article details, you must supply either an`identifier` or an `interval`.")

        if identifier:
            return [
                RxivPublicationDetail(article)
                for article in self._fetch_by_identifier(
                    server=search_server, content="pubs", identifier=identifier, format=format
                ).get("collection", [])
            ]

        if interval:
            return [
                RxivPublicationDetail(article)
                for article in self._fetch_by_interval(
                    server=search_server, content="pubs", interval=interval, cursor=cursor, format=format
                ).get("collection", [])
            ]

        return []

    def article_metadata(
        self,
        interval: tuple[str, str],
        cursor: int = 0,
        format: str = "json",
    ) -> list[RxivPublication]:
        """
        Fetch article detail for a published article matching the passed
        `interval`. This is generally for already published articles.

        The `interval` parameter can be used to filter content so that
        it matches the provided interval. Valid intervals are:
        - A tuple containing two YYYY-MM-DD dates. Note that the earlier of the
          two dates should appear first.

        The `cursor` parameter can be used to move the start cursor for the search.
        For instance, passing the number '5' will show matching results starting
        from the 5th piece of content. Default: 0

        The `format` parameter can be used to specify the return format of this
        method. Accepted values are `json` and `csv`.
        """
        return [
            RxivPublication(article)
            for article in self._fetch_by_interval(
                server=None, content="pub", interval=interval, cursor=cursor, format=format
            ).get("collection", [])
        ]

    def publisher_detail(self, prefix: str, interval: tuple[str, str], cursor: int = 0) -> list[RxivPublisherDetail]:
        """
        Fetch article details for all article matching the passed `prefix` for
        a given publisher. This is for already published articles.

        The `prefix` parameter can be used to specify the publisher prefix, which
        is the publisher prefix string prior to any slash. eg '10.15252'.

        The optional `interval` parameter can be used to filter content so that
        it matches the provided interval. Valid intervals are:
        - A tuple containing two YYYY-MM-DD dates. Note that the earlier of the
          two dates should appear first.

        The `cursor` parameter can be used to move the start cursor for the search.
        For instance, passing the number '5' will show matching results starting
        from the 5th piece of content. Default: 0
        """
        return [
            RxivPublisherDetail(publisher)
            for publisher in self._fetch_by_publisher(
                content="publisher", prefix=prefix, interval=interval, cursor=cursor
            ).get("collection", [])
        ]

    def content_statistics(self, interval: str = "m", format: str = "json") -> list[RxivContentStatistics]:
        """
        Fetch summary statistics for monthly or yearly new and revised paper
        interval and cumulative counts.

        The `interval` parameter can be used to filter content so that
        it matches the provided interval. Valid intervals are:
        - A string matching either 'm' (monthly) or 'y' (yearly).

        The `format` parameter can be used to specify the return format of this
        method. Accepted values are `json` and `csv`.
        """
        return [
            RxivContentStatistics(stats)
            for stats in self._fetch_summary_stats(content="sum", interval=interval, format=format).get(
                "collection", []
            )
        ]

    def usage_statistics(self, interval: str = "m", format: str = "json") -> list[RxivUsageStatistics]:
        """
        Fetch summary statistics for monthly or yearly abstract views, full
        text views, and PDF downloads.

        The `interval` parameter can be used to filter content so that
        it matches the provided interval. Valid intervals are:
        - A string matching either 'm' (monthly) or 'y' (yearly).

        The `format` parameter can be used to specify the return format of this
        method. Accepted values are `json` and `csv`.
        """
        return [
            RxivUsageStatistics(stats)
            for stats in self._fetch_summary_stats(content="usage", interval=interval, format=format).get(
                "collection", []
            )
        ]

    # TODO: support different interval types.
    def _fetch_by_interval(
        self,
        content: str,
        server: Optional[str],
        interval: tuple[str, str],
        cursor: int,
        format: str,
    ) -> Any:
        """
        Fetch the passed content from the provided rxiv server via the provided interval.
        """
        if server:
            return self._fetch(
                f"{self.api_url}/{content}/{server}/{interval[0]}/{interval[1]}/{cursor}/{format}", return_format=format
            )
        else:
            return self._fetch(
                f"{self.api_url}/{content}/{interval[0]}/{interval[1]}/{cursor}/{format}", return_format=format
            )

    def _fetch_by_identifier(self, server: str, content: str, identifier: str, format: str) -> Any:
        """
        Fetch the passed content from the provided Rxiv server via its identifier.
        """
        return self._fetch(f"{self.api_url}/{content}/{server}/10.1101/{identifier}/na/{format}", return_format=format)

    def _fetch_by_publisher(self, content: str, prefix: str, interval: tuple[str, str], cursor: int) -> Any:
        """
        Fetch article details from the base rxiv server via its publisher prefix.
        """
        return self._fetch(f"{self.api_url}/{content}/{prefix}/{interval[0]}/{interval[1]}/{cursor}")

    def _fetch_summary_stats(self, content: str, interval: str, format: str) -> Any:
        """
        Fetch summary statistics from the base rxiv server based on a passed interval.
        """
        return self._fetch(f"{self.api_url}/{content}/{interval}/{format}")

    def _fetch(self, url: str, return_format: str = "json") -> Any:
        """
        Fetch content from the provided url, verify the request was successful,
        and load as JSON if desired.
        """
        response = requests.get(url)

        if response.status_code == 404:
            return []

        response.raise_for_status()
        return json.loads(response.text) if return_format == "json" else response.text



"""
Utilities for managing views via SQLAlchemy.
"""

from functools import partial

import sqlalchemy as sa
from sqlalchemy.ext import compiler
from sqlalchemy.schema import DDLElement, MetaData
from sqlalchemy.orm import Session

from mavedb.db.base import Base

# See: https://github.com/sqlalchemy/sqlalchemy/wiki/Views, https://github.com/jeffwidman/sqlalchemy-postgresql-materialized-views?tab=readme-ov-file


class CreateView(DDLElement):
    def __init__(self, name: str, selectable: sa.Select, materialized: bool):
        self.name = name
        self.selectable = selectable
        self.materialized = materialized


class DropView(DDLElement):
    def __init__(self, name: str, materialized: bool):
        self.name = name
        self.materialized = materialized


class MaterializedView(Base):
    __abstract__ = True

    @classmethod
    def refresh(cls, connection, concurrently=True):
        """Refresh this materialized view."""
        refresh_mat_view(connection, cls.__table__.fullname, concurrently)


@compiler.compiles(CreateView)
def _create_view(element: CreateView, compiler, **kw):
    return "CREATE %s %s AS %s" % (
        "MATERIALIZED VIEW" if element.materialized else "VIEW",
        element.name,
        compiler.sql_compiler.process(element.selectable, literal_binds=True),
    )


@compiler.compiles(DropView)
def _drop_view(element: DropView, compiler, **kw):
    return "DROP %s %s" % ("MATERIALIZED VIEW" if element.materialized else "VIEW", element.name)


def view_exists(ddl: CreateView, target, connection: sa.Connection, materialized: bool, **kw):
    inspector = sa.inspect(connection)
    if inspector is None:
        return False

    view_names = inspector.get_materialized_view_names() if ddl.materialized else inspector.get_view_names()
    return ddl.name in view_names


def view_doesnt_exist(ddl: CreateView, target, connection: sa.Connection, materialized: bool, **kw):
    return not view_exists(ddl, target, connection, materialized, **kw)


def view(name: str, selectable: sa.Select, metadata: MetaData = Base.metadata, materialized: bool = False):
    """
    Register a view or materialized view to SQLAlchemy. Use this function to define a view on some arbitrary
    model class.

    ```
    class MyView(Base):
        __table__ = view(
            "my_view",
            select(
                MyModel.id.label("id"),
                MyModel.name.label("name"),
            ),
            materialized=False,
        )
    ```

    When registered in this manner, SQLAlchemy will create and destroy the view along with other tables. You can
    then query this view as if it were an ORM object.

    ```
    results = db.query(select(MyView.col1).where(MyView.col2)).all()
    ```
    """
    t = sa.table(
        name,
        *(sa.Column(c.name, c.type, primary_key=c.primary_key) for c in selectable.selected_columns),
    )
    t.primary_key.update(c for c in t.c if c.primary_key)  # type: ignore

    # TODO: Figure out indices.
    if materialized:
        sa.event.listen(
            metadata,
            "after_create",
            CreateView(name, selectable, True).execute_if(callable_=partial(view_doesnt_exist, materialized=True)),
        )
        sa.event.listen(
            metadata,
            "before_drop",
            DropView(name, True).execute_if(callable_=partial(view_exists, materialized=True)),
        )

    else:
        sa.event.listen(
            metadata,
            "after_create",
            CreateView(name, selectable, False).execute_if(callable_=partial(view_doesnt_exist, materialized=False)),
        )
        sa.event.listen(
            metadata,
            "before_drop",
            DropView(name, False).execute_if(callable_=partial(view_exists, materialized=False)),
        )

    return t


def refresh_mat_view(session: Session, name: str, concurrently=True):
    """
    Refreshes a single materialized view, given by `name`.
    """
    # since session.execute() bypasses autoflush, must manually flush in order
    # to include newly-created/modified objects in the refresh
    session.flush()
    _con = "CONCURRENTLY " if concurrently else ""
    session.execute(sa.text("REFRESH MATERIALIZED VIEW " + _con + name))


def refresh_all_mat_views(session: Session, concurrently=True):
    """
    Refreshes all materialized views. Views are refreshed in non-deterministic order,
    so view definitions can't depend on each other.
    """
    inspector = sa.inspect(session.connection())

    if not inspector:
        return

    for mv in inspector.get_materialized_view_names():
        refresh_mat_view(session, mv, concurrently)



import logging
from operator import attrgetter
from typing import Any

from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy.orm import Session

from mavedb import deps
from mavedb.lib.authentication import UserData, get_current_user
from mavedb.lib.logging import LoggedRoute
from mavedb.lib.logging.context import logging_context, save_to_logging_context
from mavedb.lib.permissions import Action, has_permission
from mavedb.models.experiment_set import ExperimentSet
from mavedb.view_models import experiment_set

router = APIRouter(
    prefix="/api/v1/experiment-sets",
    tags=["experiment-sets"],
    responses={404: {"description": "Not found"}},
    route_class=LoggedRoute,
)

logger = logging.getLogger(__name__)


@router.get(
    "/{urn}",
    status_code=200,
    response_model=experiment_set.ExperimentSet,
    responses={404: {}},
)
def fetch_experiment_set(
    *, urn: str, db: Session = Depends(deps.get_db), user_data: UserData = Depends(get_current_user)
) -> Any:
    """
    Fetch a single experiment set by URN.
    """
    # item = db.query(ExperimentSet).filter(ExperimentSet.urn == urn).filter(ExperimentSet.private.is_(False)).first()
    item = db.query(ExperimentSet).filter(ExperimentSet.urn == urn).first()
    save_to_logging_context({"requested_resource": urn})

    if not item:
        # the exception is raised, not returned - you will get a validation
        # error otherwise.
        logger.debug(msg="The requested resources does not exist.", extra=logging_context())
        raise HTTPException(status_code=404, detail=f"Experiment set with URN {urn} not found")
    else:
        item.experiments.sort(key=attrgetter("urn"))

    has_permission(user_data, item, Action.READ)

    # Filter experiment sub-resources to only those experiments readable by the requesting user.
    item.experiments[:] = [exp for exp in item.experiments if has_permission(user_data, exp, Action.READ).permitted]

    return item



from datetime import date
from typing import TYPE_CHECKING, List

from sqlalchemy import Boolean, Column, Date, ForeignKey, Integer, String
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import Mapped, relationship
from sqlalchemy.schema import Table

from mavedb.db.base import Base
from mavedb.lib.temp_urns import generate_temp_urn

from .contributor import Contributor
from .doi_identifier import DoiIdentifier
from .legacy_keyword import LegacyKeyword
from .publication_identifier import PublicationIdentifier
from .raw_read_identifier import RawReadIdentifier
from .user import User

if TYPE_CHECKING:
    from mavedb.models.experiment import Experiment

experiment_sets_contributors_association_table = Table(
    "experiment_set_contributors",
    Base.metadata,
    Column("experiment_set_id", ForeignKey("experiment_sets.id"), primary_key=True),
    Column("contributor_id", ForeignKey("contributors.id"), primary_key=True),
)


experiment_sets_doi_identifiers_association_table = Table(
    "experiment_set_doi_identifiers",
    Base.metadata,
    Column("experiment_set_id", ForeignKey("experiment_sets.id"), primary_key=True),
    Column("doi_identifier_id", ForeignKey("doi_identifiers.id"), primary_key=True),
)

experiment_sets_legacy_keywords_association_table = Table(
    "experiment_set_keywords",
    Base.metadata,
    Column("experiment_set_id", ForeignKey("experiment_sets.id"), primary_key=True),
    Column("keyword_id", ForeignKey("keywords.id"), primary_key=True),
)

experiment_sets_publication_identifiers_association_table = Table(
    "experiment_set_publication_identifiers",
    Base.metadata,
    Column("experiment_set_id", ForeignKey("experiment_sets.id"), primary_key=True),
    Column("publication_identifier_id", ForeignKey("publication_identifiers.id"), primary_key=True),
)

# experiment_sets_sra_identifiers_association_table = Table(
experiment_sets_raw_read_identifiers_association_table = Table(
    "experiment_set_sra_identifiers",
    Base.metadata,
    Column("experiment_set_id", ForeignKey("experiment_sets.id"), primary_key=True),
    Column("sra_identifier_id", ForeignKey("sra_identifiers.id"), primary_key=True),
)


class ExperimentSet(Base):
    __tablename__ = "experiment_sets"

    id = Column(Integer, primary_key=True)

    urn = Column(String(64), nullable=True, default=generate_temp_urn, unique=True, index=True)
    extra_metadata = Column(JSONB, nullable=False)

    private = Column(Boolean, nullable=False, default=True)
    approved = Column(Boolean, nullable=False, default=False)
    published_date = Column(Date, nullable=True)
    processing_state = Column(String, nullable=True)

    # TODO Refactor the way we handle child collections?
    num_experiments = Column(Integer, nullable=False, default=0)
    experiments: Mapped[List["Experiment"]] = relationship(
        back_populates="experiment_set", cascade="all, delete-orphan"
    )

    created_by_id = Column(Integer, ForeignKey("users.id"), index=True, nullable=True)
    created_by: Mapped[User] = relationship("User", foreign_keys="ExperimentSet.created_by_id")
    modified_by_id = Column(Integer, ForeignKey("users.id"), index=True, nullable=True)
    modified_by: Mapped[User] = relationship("User", foreign_keys="ExperimentSet.modified_by_id")
    creation_date = Column(Date, nullable=False, default=date.today)
    modification_date = Column(Date, nullable=False, default=date.today, onupdate=date.today)

    keyword_objs: Mapped[list[LegacyKeyword]] = relationship(
        "LegacyKeyword", secondary=experiment_sets_legacy_keywords_association_table, backref="experiment_sets"
    )
    contributors: Mapped[list["Contributor"]] = relationship(
        "Contributor", secondary=experiment_sets_contributors_association_table, backref="experiment_sets"
    )
    doi_identifiers: Mapped[list[DoiIdentifier]] = relationship(
        "DoiIdentifier", secondary=experiment_sets_doi_identifiers_association_table, backref="experiment_sets"
    )
    publication_identifiers: Mapped[list[PublicationIdentifier]] = relationship(
        "PublicationIdentifier",
        secondary=experiment_sets_publication_identifiers_association_table,
        backref="experiment_sets",
    )
    # sra_identifiers = relationship('SraIdentifier', secondary=experiment_sets_sra_identifiers_association_table, backref='experiment_sets')
    raw_read_identifiers: Mapped[list[RawReadIdentifier]] = relationship(
        "RawReadIdentifier", secondary=experiment_sets_raw_read_identifiers_association_table, backref="experiment_sets"
    )



from typing import Any

from pydantic import validator
from pydantic.utils import GetterDict

from mavedb.models.enums.contribution_role import ContributionRole


class PublicationIdentifiersGetter(GetterDict):
    """
    Custom class used in transforming PublicationAssociation SQLAlchemy model objects
    into Pydantic view model objects, with special handling of publication identifier
    association objects.

    Pydantic uses GetterDict objects to access source objects as dictionaries, which can
    then be turned into Pydantic view model objects. We need to remap the underlying
    SQLAlchemy model's AssociationList objects with information about whether a
    publication is primary or not into two separate lists with that same information.
    """

    def get(self, key: Any, default: Any = ...) -> Any:
        if key == "secondary_publication_identifiers":
            pub_assc = getattr(self._obj, "publication_identifier_associations")
            return [assc.publication for assc in pub_assc if not assc.primary]
        elif key == "primary_publication_identifiers":
            pub_assc = getattr(self._obj, "publication_identifier_associations")
            return [assc.publication for assc in pub_assc if assc.primary]
        else:
            return super().get(key, default)


class UserContributionRoleGetter(GetterDict):
    """
    Custom class used in transforming ContributionAssociation SQLAlchemy model objects
    into Pydantic view model objects, with special handling of user association objects.

    Pydantic uses GetterDict objects to access source objects as dictionaries, which can
    then be turned into Pydantic view model objects. We need to remap the underlying
    SQLAlchemy model's AssociationList objects with information about the role for a
    contributing user
    """

    def get(self, key: Any, default: Any = ...) -> Any:
        # The standard is to name properties as the plural of the enum value
        if key[:-1] in ContributionRole._member_map_:
            user_assc = getattr(self._obj, "user_associations")
            return [user.user for user in user_assc if key[:-1] == user.contribution_role.name]
        else:
            return super().get(key, default)


def record_type_validator():
    return validator("record_type", allow_reuse=True, pre=True, always=True)


def set_record_type(cls, v):
    # Record type will be set to the class name no matter the input.
    return cls.__name__



import time
from typing import Callable

from fastapi import Request, Response
from fastapi.routing import APIRoute
from starlette.background import BackgroundTask, BackgroundTasks

from mavedb.lib.logging.canonical import log_request
from mavedb.lib.logging.context import save_to_logging_context


class LoggedRoute(APIRoute):
    def get_route_handler(self) -> Callable:
        original_route_handler = super().get_route_handler()

        async def logging_route_handler(request: Request) -> Response:
            save_to_logging_context({"time_ns": time.time_ns()})
            response = await original_route_handler(request)
            end = time.time_ns()

            # Add logging task to list of background tasks to be completed after the response is sent.
            old_background = response.background

            task = BackgroundTask(log_request, request, response, end)
            if old_background:
                if isinstance(old_background, BackgroundTasks):
                    old_background.add_task(task)
                    response.background = old_background
                else:
                    tasks = BackgroundTasks()
                    tasks.add_task(old_background)
                    tasks.add_task(task)
                    response.background = tasks
            else:
                response.background = task

            return response

        return logging_route_handler
